from __future__ import annotations #  for 3.9 compatability

from typing import Self, Iterator

import archnemesis.database.wrappers.hapi as hapi

from .gas_descriptor import RadtranGasDescriptor, HitranGasDescriptor
from ..mappings.hitran import hitran_to_radtran
from archnemesis import Data

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.DEBUG)

class GasIsotopes:
    def __init__(self, rt_gas_id : int = None, rt_iso_ids : tuple[int,...] = tuple()):
        """
        Stored using radtran ID mappings internally
        """
        if rt_gas_id is not None:
            if type(rt_iso_ids) is int:
                if rt_iso_ids == 0:
                    rt_iso_ids = tuple(i+1 for i in range(Data.gas_data.count_isotopes(rt_gas_id)))
                else:
                    rt_iso_ids = (rt_iso_ids,)
            else:
                rt_iso_ids = tuple(x for x in rt_iso_ids)
                
        for x in rt_iso_ids:
            if x == 0:
                raise ValueError('Gas isotope id should not be 0, should be split up into all isotopes e.g (1,2,3,4,...)')
    
        self.rt_gas_id : None | int = rt_gas_id
        self.rt_iso_ids : tuple[int] = rt_iso_ids
    
    def __repr__(self):
        return f'GasIsotopes({self.rt_gas_id=}, {self.rt_iso_ids=})'
    
    
    @classmethod
    def from_radtran(cls, rt_gas_id : int, rt_iso_ids : int | tuple[int,...]) -> Self:
        return cls(rt_gas_id, rt_iso_ids)
    
    @classmethod
    def from_radtran_gasses(cls, gas_descs : tuple[RadtranGasDescriptor,...]) -> Self:
        gas_ids = tuple(set([x.gas_id for x in gas_descs]))
        assert len(gas_ids) == 1, "When creating GasIsotopes instance, must supply multiple isotopes of the same gas"
        iso_ids = tuple(sorted([x.iso_id for x in gas_descs]))
        return cls(gas_ids[0], iso_ids)
    
    @classmethod
    def from_hitran(cls, ht_gas_id : int, ht_iso_id : int | tuple[int,...]) -> Self:
        if type(ht_gas_id ) is int:
            if ht_iso_id == 0:
                n_hitran_isotopes = len(True for x in hapi.ISO.keys() if x[0]==ht_gas_id)
                ht_iso_id = tuple(i+1 for i in range(n_hitran_isotopes))
            else:
                ht_iso_id = tuple(ht_iso_id,)
        else:
            ht_iso_id = tuple(x for x in ht_iso_id)
        
        rt_gas_id_isos = []
        for iso_id in ht_iso_id:
            id, iso = hitran_to_radtran.get((ht_gas_id, iso_id),(None,None))
            if id is not None:
                rt_gas_id_isos.append((id, iso))
        
        rt_gas_id = rt_gas_id_isos[0][0]
        rt_iso_id = tuple(x[1] for x in rt_gas_id_isos)
        
        assert all((x[0] == rt_gas_id for x in rt_gas_id_isos)), "When creating a GasIsotopesDescriptor from HITRAN ID values, all gasses must map to the same RADTRAN ID values"
        
        return cls(rt_gas_id, rt_iso_id)
    
    @property
    def n_isotopes(self) -> int:
        return len(self.rt_iso_ids)
    
    @property
    def gas_name(self) -> str:
        return Data.gas_info[str(self.rt_gas_id)]['name']
    
    @property
    def iso_names(self) -> Iterator[str,...]:
        return (Data.gas_info[str(self.rt_gas_id)]['isotope'][str(iso_id)]['name'] for iso_id in self.rt_iso_ids)
    
    @property
    def is_valid(self) -> bool:
        return self.rt_gas_id is not None and len(self.rt_iso_ids) > 0
    
    def as_hitran(self) -> tuple[int, tuple[int,...], tuple[int,...]]:
        g = tuple(self.as_hitran_gasses())
        ht_gas_id = [x.gas_id for x in g if x is not None][0]
        ht_iso_ids = [x.iso_id if x is not None else None for x in g]
        ht_global_ids = [x.global_id if x is not None else None for x in g]
        return ht_gas_id, ht_iso_ids, ht_global_ids
    
    def as_radtran_gasses(self) -> Iterator[RadtranGasDescriptor, ...]:
        return (RadtranGasDescriptor(self.rt_gas_id, iso_id) for iso_id in self.rt_iso_ids)

    def as_hitran_gasses(self, filter_missing=True) -> Iterator[None|HitranGasDescriptor, ...]:
        if filter_missing:
            return (x.to_hitran() for x in self.as_radtran_gasses() if x is not None)
        else:
            return (x.to_hitran() for x in self.as_radtran_gasses())
    
    def contains(self, other : GasIsotopes) -> bool:
        """
        Does this set of gas isotopes contain `other`
        """
        if self.rt_gas_id != other.rt_gas_id:
            return False
        for iso_id in other.rt_iso_ids:
            if iso_id not in self.rt_iso_ids:
                return False
        
        return True


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
#
# archNEMESIS - Python implementation of the NEMESIS radiative transfer and retrieval code
# Multiple_Scattering_Core.py - Core functions for multiple scattering calculations.
#
# Copyright (C) 2025 Juan Alday, Joseph Penn, Patrick Irwin,
# Jack Dobinson, Jon Mason, Jingxuan Yang
#
# This file is part of archNEMESIS.
#
# archNEMESIS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from numba import njit#, prange


# @njit(fastmath=True, error_model='numpy')
# def phase1(calpha, iscat, cons, icons=0, icont=None, ncont=None, vwave=None, pfunc = None, xmu = None):
#     pi = np.pi
#     calpha = min(max(calpha, -1.0), 1.0)
#     if iscat == 0:
#         p = 0.75 * (1.0 + calpha * calpha)
#     elif iscat == 1:
#         p = 1.0
#     elif iscat == 2:
#         f1 = cons[0]
#         f2 = 1.0 - f1
#         hg11 = 1.0 - cons[1] * cons[1]
#         hg12 = 2.0 - hg11
#         hg21 = 1.0 - cons[2] * cons[2]
#         hg22 = 2.0 - hg21
#         p = f1 * hg11 / np.sqrt(hg12 - 2.0 * cons[1] * calpha) ** 3 + f2 * hg21 / np.sqrt(hg22 - 2.0 * cons[2] * calpha) ** 3
#     elif iscat == 3:
#         cons[0] = 0.25 / pi
#         p = 0.0
#         xf = calpha * calpha
#         for k in range(1, (icons + 1) // 2 + 1):
#             n0 = 2 * k - 1
#             n1 = 2 * k
#             x0 = 1.0
#             x1 = calpha
#             pa0 = 0.0
#             pa1 = 0.0
#             for i in range(1, k + 1):
#                 pa0 += x0 * jcoef[i - 1, n0 - 1] / jdiv[k - 1]
#                 pa1 += x1 * jcoef[i - 1, n1 - 1] / jdiv[k - 1]
#                 x0 *= xf
#                 x1 *= xf
#             p += cons[n0] * pa0 + cons[n1] * pa1
#     elif iscat == 4:
#         p = np.interp(calpha,xmu,pfunc)
        
#     else:
#         print('Error invalid scattering option.')
#         return None
#     p /= 4.0 * pi
#     return p

@njit
def phasint2_ref(nphi, ic, nmu, mu, iscat, cons, ncons, icont, ncont, pfunc, xmu):
    '''
    Function to compute the phase matrix for a Hapke surface
    '''
    
    pi = np.pi
    dphi = 2.0 * pi / nphi
    ntheta = nmu*nmu*(nphi+1)
    
    #Defining scattering angles
    ix = 0
    cpl = np.zeros(ntheta)
    cmi = np.zeros(ntheta)
    for j in range(nmu):
        for i in range(nmu):
            sthi = np.sqrt(1.0-mu[i]*mu[i]) #sin(theta[i])
            sthj = np.sqrt(1.0-mu[j]*mu[j]) #sin(theta[i])
            for k in range(nphi+1):
                phi = k*dphi
                
                #Calculating cos(alpha)
                cpl[ix] = sthi*sthj*np.cos(phi) + mu[i]*mu[j]
                cmi[ix] = sthi*sthj*np.cos(phi) - mu[i]*mu[j]
                
                if(cpl[ix]>1.):
                    cpl[ix] = 1.
                if(cpl[ix]<-1.):
                    cpl[ix] = -1.
                if(cmi[ix]>1.):
                    cmi[ix] = 1.
                if(cmi[ix]<-1.):
                    cmi[ix] = -1.
                    
                ix += 1
        
    #Calculating the phase function at the scattering angles
    if iscat == 0: #Rayleigh scattering
        ppl = 0.75 * (1.0 + cpl * cpl)/ (4 * np.pi)
        pmi = 0.75 * (1.0 + cmi * cmi)/ (4 * np.pi)
    else:
        ppl = np.interp(cpl, xmu, pfunc) #phase1(cpl, iscat, cons, ncons, icont, ncont, vwave, pfunc, xmu)
        pmi = np.interp(cmi, xmu, pfunc) #phase1(cmi, iscat, cons, ncons, icont, ncont, vwave, pfunc, xmu)        
        
    #Integrating the phase function over the azimuth direction
    pplpl = np.zeros((nmu,nmu))
    pplmi = np.zeros((nmu,nmu))
    ix = 0
    for j in range(nmu):
        for i in range(nmu):
            for k in range(nphi+1):
                phi = k*dphi
                    
                plx = ppl[ix] * np.cos(ic*phi)
                pmx = pmi[ix] * np.cos(ic*phi)
                
                wphi = 1.*dphi
                if k == 0:
                    wphi = 0.5*dphi
                elif k == nphi:
                    wphi = 0.5*dphi
                
                if ic == 0:
                    wphi = wphi / (2.*np.pi)
                else:
                    wphi = wphi / np.pi
                
                pplpl[i,j] += wphi * plx
                pplmi[i,j] += wphi * pmx
                
                ix += 1
                
    return (pplpl, pplmi)
                

@njit
def phasint2(nphi, ic, nmu, mu, iscat, cons, ncons, icont, ncont, pfunc, xmu):
    
    pi = np.pi
    dphi = 2.0 * pi / nphi

    phi = np.arange(nphi + 1) * dphi
    sth = np.sqrt(1.0 - mu**2)
    mu2d = mu[:, None]
    
    cos_ic_phi = np.cos(ic * phi)
    cos_phi_grid = np.cos(phi)

    # Broadcasting
    sth_sth = sth[:, None] * sth[None, :]
    mu_mu = mu2d * mu2d.T
    
    cpl = sth_sth[:, :, None] * cos_phi_grid + mu_mu[:, :, None]
    cmi = sth_sth[:, :, None] * cos_phi_grid - mu_mu[:, :, None]
    #Cutting out phase1 for now

    if iscat == 0:
        pl = 0.75 * (1.0 + cpl * cpl)/ (4 * np.pi)
        pm = 0.75 * (1.0 + cmi * cmi)/ (4 * np.pi)
    elif iscat == 2:
        f1 = pfunc[0]
        f2 = 1.0 - f1
        hg11 = 1.0 - pfunc[1] * pfunc[1]
        hg12 = 2.0 - hg11
        hg21 = 1.0 - pfunc[2] * pfunc[2]
        hg22 = 2.0 - hg21
        pl = f1 * hg11 / np.sqrt(hg12 - 2.0 * pfunc[1] * cpl) ** 3 + f2 * hg21 / np.sqrt(hg22 - 2.0 * pfunc[2] * cpl) ** 3
        pm = f1 * hg11 / np.sqrt(hg12 - 2.0 * pfunc[1] * cmi) ** 3 + f2 * hg21 / np.sqrt(hg22 - 2.0 * pfunc[2] * cmi) ** 3
        pl /= 4*np.pi
        pm /= 4*np.pi
    else:
        pl = np.interp(cpl, xmu, pfunc) #phase1(cpl, iscat, cons, ncons, icont, ncont, vwave, pfunc, xmu)
        pm = np.interp(cmi, xmu, pfunc) #phase1(cmi, iscat, cons, ncons, icont, ncont, vwave, pfunc, xmu)        
        
    wphi = np.full(phi.shape, dphi)
    wphi[0]  = 0.5 * dphi
    wphi[-1] = 0.5 * dphi
    
    if ic == 0:
        wphi /= (2.0 * pi)
    else:
        wphi /= pi
    
    wphi = wphi[None, None, :]
    
    plx = pl * cos_ic_phi
    pmx = pm * cos_ic_phi
    
    pplpl = np.sum(wphi * plx, axis=2)
    pplmi = np.sum(wphi * pmx, axis=2)

    return (pplpl, pplmi)


@njit(fastmath=True, error_model='numpy')
def hansen(ic, ppl, pmi, wtmu, nmu, fc):
    '''
    Subroutine to normalise the phase function using the method described in Hansen (1971,J.ATM.SCI., V28, 1400)

    PPL,PMI are the forward and backward parts of the azimuthally-integrated
    phase function. The normalization of the true phase fcn is:
    integral over sphere [ P(mu,mu',phi) * dO] = 1
    where dO os the element of solid angle and phi is the azimuthal angle.
    '''
    
    pi = np.pi
    x1 = 2.0 * pi
    if ic == 0:
        rsum = np.zeros(nmu, dtype=float)
        for j in range(nmu):
            rsum[j] = np.sum(pmi[:, j] * wtmu) * x1
        niter = 0
        tsum = np.zeros(nmu)
        
        for niter in range(10000):
            for j in range(nmu):
                tsum[j] = np.sum(ppl[:, j] * wtmu * fc[:, j]) * x1
            testj = np.abs(rsum + tsum - 1.0)
            test = np.max(testj)
            if test < 1e-14:
                break
            for j in range(nmu):
                xj = (1.0 - rsum[j]) / tsum[j]
                for i in range(j+1):
                    xi = (1.0 - rsum[i]) / tsum[i]
                    fc[i, j] = 0.5 * (fc[i, j] * xj + fc[j, i] * xi)
                    fc[j, i] = fc[i, j]
    ppl *= fc
    return (ppl, fc)

@njit(fastmath=True)
def calc_pmat6(ic, mu, wtmu, nmu, iscat, cons8, ncons, norm, icont, ncont, nphi, fc, pfunc, xmu):
    
    #Integrating the phase function
    pplpl, pplmi = phasint2(nphi, ic, nmu, mu, iscat, cons8, ncons, icont, ncont, pfunc, xmu)

    if norm == 1:
        pplpl, fc = hansen(ic, pplpl, pplmi, wtmu, nmu, fc)

    return (pplpl, pplmi, fc)
            
@njit(fastmath=True)
def matmul(A, B):
    '''
    Matrix multiplication
    '''
    # Get the dimensions of the matrices
    rows_A, cols_A = A.shape
    rows_B, cols_B = B.shape

    # Ensure the matrices are compatible for multiplication
    if cols_A != rows_B:
        raise ValueError("Number of columns in A must be equal to number of rows in B")

    # Initialize the result matrix with zeros
    result = np.zeros((rows_A, cols_B))

    # Perform matrix multiplication using nested loops
    for i in range(rows_A):
        for j in range(cols_B):
            for k in range(cols_A):
                result[i, j] += A[i, k] * B[k, j]

    return result

@njit
def frob(r):
    return np.sqrt((r*r).sum())

@njit(fastmath=True)
def add(r1, t1, j1, e, nmu,ic):

    bcom = matmul(r1,r1)
    if frob(r1)>0.1: # Frobenius norm < 0.1 for approximation to keep error under 1e-4
        acom = np.linalg.inv(e-bcom)
    else: 
        acom = e + bcom
    ccom = matmul(t1,acom)
    rans = matmul(ccom,r1)
    acom = matmul(rans,t1)
    rans = r1 + acom
    tans = matmul(ccom,t1)

    if ic==0:
        jcom = matmul(r1,j1)
        jcom = jcom + j1
        jans = matmul(ccom,jcom)
        jans = jans + j1
    else:
        jans = j1

    return rans, tans, jans


@njit(fastmath=True)
def numba_diagonal(arr):
    rows, cols = arr.shape
    length = min(rows, cols)
    diagonal = np.empty(length, arr.dtype)
    for i in range(length):
        diagonal[i] = arr[i, i]
    return diagonal

@njit(fastmath=True)
def numba_fill_diagonal(arr, vec):
    for i in range(len(vec)):
        arr[i, i] = vec[i]
        
@njit(fastmath = True)
def numba_sum_diagonal(arr):
    sum_ = 0
    for i in range(arr.shape[0]):
        sum_ += arr[i,i]
    return sum_

@njit(fastmath = True, error_model='numpy')
def double1(ic,nmu,cc,pplpl,pplmi,omega,taut,bc,mminv,e,raman = False): 
    
    ipow0 = 12

    con = omega * np.pi
    del01 = 0.0
    if ic == 0:
        del01 = 1.0
    
    con *= (1.0 + del01)
    
    #Computation of Gamma++ (Plass et al. 1973)
    #GPLPL = MMINV*(E - CON*PPLPL*CC)
    acom = matmul(pplpl,cc)*con
    bcom = e - acom
    gplpl = matmul(mminv,bcom)
    
    #Computation of Gamma+- (Plass et al. 1973)
    #GPLMI = MMINV*CON*PPLMI*CC
    acom = matmul(pplmi,cc)*con
    gplmi = matmul(mminv,acom)
    
    nn = int(np.log2(taut) + ipow0)
    xfac = 1.0 / (2.0 ** nn) if nn >= 1 else 1.0
    tau0 = taut * xfac
    
    # Computation of R, T and J for initial layer
    t1 = e - tau0 * gplpl
    r1 = tau0 * matmul(e,gplmi)
    
    j1 = np.zeros((nmu,1))
    if ic == 0:
        for i in range(nmu):
            j1[i] = (1.0 - omega) * bc * tau0 * mminv[i,i]

    if nn < 1:
        return r1, t1, j1
    
    for n in range(nn):
        r1, t1, j1 = add(r1, t1, j1, e, nmu, ic)
   
    return r1, t1, j1

@njit(fastmath = True)
def idown(ra,ta,ja,rb,tb,jb,u0pl,utmi,e):
    '''
    Calculate the downward radiance at any interior point tau_1 in the layer from tau_0 and tau_2
    
    This function calculates equation 6 from Plass et al. (1973)
    
    I1+ = (E-R10*R12)^-1 * (T01*I0+ + R10*T21*I2- + J01+ + R10*J21- )
    
    Inputs
    ______
    
    ra(nmu,nmu) :: Reflection matrix R10 
    ta(nmu,nmu) :: Transmission matrix T01
    ja(nmu,nmu) :: Source matrix J01+
    rb(nmu,nmu) :: Reflection matrix R12
    tb(nmu,nmu) :: Transmission matrix T21
    jb(nmu,nmu) :: Source matrix J21-
    u0pl(nmu,1) :: Downward intensity I0+
    utmi(nmu,1) :: Upward intensity I2-
    e(nmu,nmu) :: Identity matrix
    
    Outputs
    _______
    
    upl(nmu,1) :: Downward intensity I1+
    '''
    
    #Calculate R10*R12
    acom = matmul(ra,rb)
    
    #Calculate (E-R10*R12)^-1
    bcom = np.linalg.inv(e-acom)
    
    #Calculate T01*I0+
    xcom = matmul(ta,u0pl)
    
    #Calculate R10*T21
    acom = matmul(ra,tb)

    #Calculate R10*T21*I2-
    ycom = matmul(acom,utmi)
    
    #Add previous two results (T01*I0+ + R10*T21*I2-)
    xcom += ycom

    #Calculate R10*J21-
    ycom = matmul(ra,jb)
    
    #Add to total (T01*I0+ + R10*T21*I2- + R10*J21- + J01+)
    xcom += + ycom + ja

    #Multiply by (E-R10*R12)^-1 to get final I1+
    upl = matmul(bcom,xcom)

    return upl

@njit(fastmath = True)
def iup(ra,ta,ja,rb,tb,jb,u0pl,utmi,e):
    '''
    Calculate the upward radiance at any interior point tau_1 in the layer from tau_0 and tau_2
    
    This function calculates equation 5 from Plass et al. (1973)
    
    I1- = (E-R12*R10)^-1 * (T21*I2- + R12*T01*I0+ + J21- + R12*J01+ )
    
    Inputs
    ______
    
    ra(nmu,nmu) :: Reflection matrix R10 
    ta(nmu,nmu) :: Transmission matrix T01
    ja(nmu,nmu) :: Source matrix J01+
    rb(nmu,nmu) :: Reflection matrix R12
    tb(nmu,nmu) :: Transmission matrix T21
    jb(nmu,nmu) :: Source matrix J21-
    u0pl(nmu,1) :: Downward intensity I0+
    utmi(nmu,1) :: Upward intensity I2-
    e(nmu,nmu) :: Identity matrix
    
    Outputs
    _______
    
    umi(nmu,1) :: Upward intensity I1-
    '''
    
    #Calculate R12*R10
    acom = matmul(rb,ra)
    
    #Calculate (E-R12*R10)^-1
    bcom = np.linalg.inv(e-acom)
    
    #Calculate T21*I2-
    xcom = matmul(tb,utmi)
    
    #Calculate R12*T01
    acom = matmul(rb,ta)

    #Calculate R12*T01*I0+
    ycom = matmul(acom,u0pl)
    
    #Add previous two results (T12*I2- + R12*T01*I0+)
    xcom += ycom

    #Calculate R21*J01+
    ycom = matmul(rb,ja)
    
    #Add to total (T12*I2- + R12*T01*I0+ + R21*J01+ + J21-)
    xcom += + ycom + jb

    #Multiply by (E-R12*R10)^-1 to get final I1-
    umi = matmul(bcom,xcom)

    return umi
    


@njit(fastmath = True)
def addp(r1, t1, j1, iscat1,e, rsub, tsub, jsub, nmu): 

    '''
    Calculate the the R,T,J matrices from the combination of two layers
    '''

    if iscat1 == 1:
        rsq = np.zeros_like(r1)
        bcom = np.zeros_like(r1)
        ccom = np.zeros_like(r1)
        rans = np.zeros_like(r1)
        tans = np.zeros_like(r1)
        jcom = np.zeros_like(j1)
        jans = np.zeros_like(j1)
        
        
        # Second layer is scattering
        rsq = matmul(rsub,r1)
        if frob(rsq)>0.01:
            acom = np.linalg.inv(e-rsq)
        else:
            acom = e+rsq
        ccom = matmul(t1,acom)
        rans = matmul(ccom,rsub)
        bcom = matmul(rans,t1)
        rans = r1 + bcom
        tans = matmul(ccom,tsub)
        jcom = matmul(rsub,j1)
        jcom += jsub
        jans = matmul(ccom,jcom)
        jans += j1
        
    else:
        
        # Second layer is non-scattering
        jcom = np.zeros_like(j1)
        tans = np.zeros((nmu,nmu))
        rans = np.zeros((nmu,nmu))
        jans = np.zeros((nmu,1))
        
        jcom = matmul(rsub,j1)
        jcom += jsub
        
        for i in range(nmu):
            ta = t1[i, i]
            for j in range(nmu):
                tb = t1[j, j]   
                tans[i, j] = tsub[i, j] * ta
                rans[i, j] = rsub[i, j] * ta * tb
            jans[i, 0] = j1[i, 0] + ta * jcom[i, 0]
            
    return rans, tans, jans

@njit(fastmath = True)
def angle_quadrature(solar,sol_ang,emiss_ang,mu,nmu):

    if sol_ang > 90.0:
        zmu0 = np.cos(np.radians(180 - sol_ang))
        solar1 = solar*0.0
    else:
        zmu0 = np.cos(np.radians(sol_ang))
        solar1 = solar

    zmu = np.cos(np.radians(emiss_ang))

    isol = 0
    for j in range(nmu-1):
        if zmu0 <= mu[j] and zmu0 > mu[j+1]:
            isol = j
    if zmu0 <= mu[nmu-1]:
        isol = nmu - 2

    iemm = 0
    for j in range(nmu-1):
        if zmu <= mu[j] and zmu > mu[j+1]:
            iemm = j
    if zmu <= mu[nmu-1]:
        iemm = nmu - 2

    u = (mu[isol] - zmu0) / (mu[isol] - mu[isol+1])
    t = (mu[iemm] - zmu) / (mu[iemm] - mu[iemm+1])
    
    return solar1, isol, iemm, t, u 

@njit(fastmath = True, error_model='numpy')
def calc_rtj_matrix(ic,mu,wtmu,bb,tautot,tauscat,tauray,frac,ppln,pmin,pplr,pmir,cc,mminv,e):
    '''
    Calculate the Reflection, Transmission and Source matrices for a given atmospheric layer
    
    Inputs
    ______
    
    mu(nmu) :: Quadrature angles
    wtmu(nmu) :: Weights of the quadrature angles
    bb :: Black-body radiation for the layer 
    tautot :: Total optical depth of the layer (absorption + scattering)
    omega :: Aerosol scattering optical depth of the layer
    tauray :: Rayleigh scattering optical depth of the layer
    frac(ndust) :: Fraction of aerosol scattering by each aerosol population
    ppln(ndust,nmu,nmu) :: Phase matrix in + direction of each aerosol population
    pmin(ndust,nmu,nmu) :: Phase matrix in - direction of each aerosol population
    pplr(nmu,nmu) :: Phase matrix in + direction for Rayleigh scattering
    pmir(nmu,nmu) :: Phase matrix in - direction for Rayleigh scattering
    
    Outputs
    _______
    
    rl(nmu,nmu) :: Reflection matrix
    tl(nmu,nmu) :: Transmission matrix
    jl(nmu,1) :: Source matrix
    iscl :: Flag indicating whether layer is scattering (1) or non-scattering (0)
    '''

    nmu = len(mu)       #Number of quadrature angles
    ncont = len(frac)   #Number of aerosol populations
    omega = (tauscat+tauray)/tautot  #Single scattering albedo of the layer
    
    #Initialising arrays
    rl = np.zeros((nmu,nmu))
    tl = np.zeros((nmu,nmu))
    jl = np.zeros((nmu,1))
    
    #Calculating matrices
    if cc is None:
        cc = np.zeros((nmu,nmu))
        numba_fill_diagonal(cc, wtmu)
    if e is None:
        e = np.identity(nmu)
    if mminv is None:
        mminv = np.zeros((nmu,nmu))
        numba_fill_diagonal(mminv,1./mu)
    
    #Calculating matrices
    if tautot == 0:     #If there is no atmospheric opacity
        iscl = 0
        rl[:,:] *= 0.0
        tl[:,:] *= 0.0
        jl[:,:] *= 0.0
        for i in range(nmu):
            tl[i,i] = 1.0

    elif omega == 0:  #If layer is not scattering
        iscl = 0
        rl[:,:] *= 0.0
        tl[:,:] *= 0.0
        for i1 in range(nmu):
            tex = -mminv[i1,i1]*tautot
            if tex > -200.0:
                tl[i1,i1] = np.exp(tex)
            else:
                tl[i1,i1] = 0.0

            jl[i1,0] = bb*(1.0 - tl[i1,i1])

    else:  #If layer is scattering
        pplpl = (tauray/(tauscat+tauray))*pplr[:,:]
        pplmi = (tauray/(tauscat+tauray))*pmir[:,:]

        for j1 in range(ncont):
            pplpl += tauscat/(tauscat+tauray)*ppln[j1,:,:]*frac[j1] 
            pplmi += tauscat/(tauscat+tauray)*pmin[j1,:,:]*frac[j1] 

        iscl = 1
        rl[:,:], tl[:,:], jl[:,:] = double1(ic,nmu,cc,pplpl,pplmi,omega,tautot,bb,mminv,e,raman = False)
    
    return rl,tl,jl,iscl



@njit(fastmath = True,parallel=False, cache = True, error_model='numpy')
def scloud11wave_core(phasarr, radg, sol_angs, emiss_angs, solar, aphis, lowbc, brdf_matrix, mu1, wt1, nf,
                vwaves, bnu, taus, tauray,omegas_s, nphi,iray,imie, lfrac):
    
    """
    Calculate spectrum using the adding-doubling method.

    Parameters
    ----------
    phasarr(NMODES,NWAVE,2,NPHAS):
        Contains phase functions (phasarr[:,:,0,:]) and corresponding angle grids (phasarr[:,:,1,:])
    radg(NWAVE,NMU): 
        Incident intensity at the bottom of the atm
    sol_ang(NPATH):
        Solar zenith angle (degrees)
    emiss_angs(NPATH):
        Emission zenith angle (degrees)
    solar(NWAVE):        
        Incident solar flux at the top of the atm
    aphi(NPATH):
        Azimuth angle between Sun and observer
    lowbc:
        Lower boundary condition: 0 = thermal (no surface), 1 = Lambert reflection (surface)
    brdf_matrix(NWAVE,NMU,NMU,NF+1):
        Decomposed BRDF matrix of the surface
    mu1(NMU):  
        Zenith angle point quadrature
    wt1(NMU):  
        Zenith angle point quadrature
    nf:
        Maximum number of terms in azimuth Fourier expansion 
    vwaves(NWAVE):
        Input wavelengths.
    bnu(NWAVE,NLAY):
        Mean Planck function in each layer
    tau(NWAVE,NG,NLAY):
        Total optical thickness of each layer
    tauray(NWAVE,NLAY):
        Rayleigh optical thickness of each layer
    omegas(NWAVE,NG,NLAY):
        Single scattering albedo of each layer (aerosol + Rayleigh scattering)
    nphi:
        Number of azimuth integration ordinates
    iray:
        Flag for using rayleigh scattering
    iray:
        Flag for phase function type
    lfrac(NWAVE,NCONT,NLAY):
        Fraction of scattering contributed by each type in each layer
    
    Returns
    -------
    rad(NWAVE):
        Upwards radiance at each wavelength
    """
    
    nwave = len(vwaves)
    nmu = len(mu1)
    ngeom = len(emiss_angs)
    ng = taus.shape[1]
    nlay = taus.shape[2]
    ncont = phasarr.shape[0]
    #pi = np.pi

    ltot = nlay
    if lowbc > 0:
        ltot += 1

    xfac = 0.
    xfac = np.sum(mu1*wt1)
    xfac = 0.5/xfac    

    # Reset the order of angles
    mu = mu1[::-1]
    wtmu = wt1[::-1]
    
    # Setting up radiance matrices
    rad = np.zeros((ngeom,ng,nwave))

    yx = np.zeros((4))
    
    u0pl = np.zeros((nmu,1))           #Incident downward radiance (+) at top of the atmosphere (solar)
    utmi = np.zeros((nmu,1))           #Incident upward radiance (-) at bottom of atmosphere or surface (ground)
    umi = np.zeros((nlay,nmu))         #Upward radiance (-)
    upl = np.zeros((nlay,nmu))         #Downward radiance (+)

    ppln = np.zeros((ncont, nmu, nmu)) #Phase matrix of the scattering aerosols in + direction
    pmin = np.zeros((ncont, nmu, nmu)) #Phase matrix of the scattering aerosols in - direction
    pplr = np.zeros((nmu, nmu))        #Phase matrix for Rayleigh scattering in + direction
    pmir = np.zeros((nmu, nmu))        #Phase matrix for Rayleigh scattering in - direction
    
    rl = np.zeros((nmu,nmu))           #Reflection matrix of each layer
    tl = np.zeros((nmu,nmu))           #Transmission matrix of each layer
    jl = np.zeros((nmu,1))             #Source matrix of each layer
    
    rs = np.zeros((nmu,nmu,nf+1))      #Reflection matrix of the surface
    ts = np.zeros((nmu,nmu,nf+1))      #Tranmission matrix of the surface
    js = np.zeros((nmu,1,nf+1))        #Source matrix of the surface

    rcomb = np.zeros((nmu,nmu,nf+1))   #Combined reflection matrix from multiple layers
    tcomb = np.zeros((nmu,nmu,nf+1))   #Combined transmission matrix from multiple layers
    jcomb = np.zeros((nmu,1,nf+1))     #Combined source matrix from multiple layers

    # Setting up matrices of constants 
    e = np.identity(nmu)
    mm = np.zeros((nmu, nmu))
    numba_fill_diagonal(mm, mu[:nmu])
    mminv = np.zeros((nmu, nmu))
    numba_fill_diagonal(mminv, 1./mu[:nmu])
    cc = np.zeros((nmu, nmu))
    numba_fill_diagonal(cc, wtmu)
    ccinv = np.zeros((nmu, nmu))
    numba_fill_diagonal(ccinv, 1/wtmu[:nmu])
    
    radg = radg[:,::-1]

    fc = np.ones((ncont+1,nmu,nmu))
    

    #Performing initial checks on whether all spectra are for looking up or looking down
    if np.all(emiss_angs < 90):
        lookdown = True
    elif np.all(emiss_angs > 90):
        lookdown = False
    else:
        raise ValueError("Emission angles are a mix of values above and below 90 degrees.")

        
    #Main loop: Iterating through each wavelength
    for ig in range(ng):
        # Getting correct layer properties for this g-ordinate
        
        tau = taus[:,ig,:]
        omegas = omegas_s[:,ig,:]
        
        for widx in range(nwave):
            
            # Expand into successive fourier components until convergence or ic = nf
            
            for ic in range(nf+1):
                ppln.fill(0.)
                pmin.fill(0.)
                pplr.fill(0.)
                pmir.fill(0.)

                for j1 in range(ncont):                    
                    pfunc = phasarr[j1, widx, 0, :]
                    xmu   = phasarr[j1, widx, 1, :]
                    if imie == 0:
                        iscat = 2
                    else:
                        iscat = 4
                    ncons = 0
                    cons8 = pfunc
                    norm = 1
                    pplpl, pplmi, fc[j1] = calc_pmat6(ic, mu, wtmu, nmu, iscat, cons8, ncons, 
                                                norm, j1, ncont, nphi, fc[j1], pfunc, xmu)
                    # Transfer matrices to those for each scattering particle
                    ppln[j1,:,:] = pplpl
                    pmin[j1,:,:] = pplmi

                if iray > 0:
                    iscat = 0
                    ncons = 0
                    pplpl, pplmi, fc[ncont] = calc_pmat6(ic, mu, wtmu, nmu, iscat, cons8, ncons, 
                                                1, ncont, ncont, nphi, fc[ncont], pfunc, xmu)

                    # Transfer matrices to storage
                    pplr[:,:] = pplpl
                    pmir[:,:] = pplmi
                    
                #Computing RTJ matrices for the surface
                surface_defined = False
                if(lowbc > 0):
                    
                    js[:,0,ic] = radg[widx]   #Emissivity is already accounted for in radg
                    for j in range(nmu):
                        rs[:,j,ic] = 2. * (brdf_matrix[widx,:,j,ic] * np.pi) * mu[j] * wtmu[j]
                    rs[:,:,ic]*= xfac
                    ts[:,:,ic].fill(0.0)
    
                    if lookdown == True:
                        surface_defined = True
                        jcomb[:,:,ic] = js[:,:,ic]
                        rcomb[:,:,ic] = rs[:,:,ic]
                        tcomb[:,:,ic] = ts[:,:,ic]

                # Main loop: computing R,T,J for each layer
                for l in range(0,nlay):
                    
                    if lookdown is True:
                        k = l
                    else:
                        k = nlay - 1 - l
                        
                    #Defining the properties of the layer
                    taut = tau[widx,k]
                    bc = bnu[widx,k]
                    omega = omegas[widx,k]
                    if omega < 0:
                        omega = 0.0
                    if omega > 1:
                        omega = 1.0

                    tauscat = taut*omega
                    taur = tauray[widx,k]
                    tauscat = tauscat-taur

                    if tauscat < 0:
                        tauscat = 0.0  
                        
                    omega = (tauscat+taur)/taut
                    frac = lfrac[widx,:,k]

                    #Calculating the matrices
                    rl, tl, jl, iscl = calc_rtj_matrix(ic,mu,wtmu,bc,taut,tauscat,taur,frac,ppln,pmin,pplr,pmir,cc,mminv,e)
                    
                    #Combining layers along the path
                    if l == 0 and surface_defined == False:
                        jcomb[:,:,ic] = jl
                        rcomb[:,:,ic] = rl
                        tcomb[:,:,ic] = tl
                    else:
                        rcomb[:,:,ic], tcomb[:,:,ic], jcomb[:,:,ic] = addp(rl, tl, jl, iscl,
                                                    e, rcomb[:,:,ic], tcomb[:,:,ic], jcomb[:,:,ic], nmu)

                    #print('rcomb',rcomb[:,:,ic])
                    #print('tcomb',tcomb[:,:,ic])
                    #print('jcomb',jcomb[:,:,ic])
                    #raise ValueError('hola')

                if ic != 0:
                    jcomb[:,:,ic].fill(0.0)

            #calculating the spectra
            for ipath in range(ngeom):
            
                #converged = False
                conv1 = False
                defconv = 1e-5
                
                sol_ang = sol_angs[ipath]
                emiss_ang = emiss_angs[ipath]
                aphi = aphis[ipath]
                
                if emiss_ang < 90.: #Nadir-viewing geometry
                    new_emi = emiss_ang
                elif emiss_ang > 90.: #Upward-looking geometry
                    new_emi = 180. - emiss_ang
                    
                solar1, isol, iemm, t, u = angle_quadrature(solar,sol_ang,new_emi,mu,nmu)
            
                for ic in range(nf+1):
                    
                    #Calculating the spectrum
                    for j in range(nmu):
                        u0pl[j] = 0.0
                        if ic == 0:
                            utmi[j] = radg[widx,j]
                        else:
                            utmi[j] = 0.0

                    ico = 0

                    if lookdown is True:
                        for imu0 in range(isol, isol+2): 
                            u0pl[imu0] = solar1[widx] / (2.0 * np.pi * wtmu[imu0])
                            acom = matmul(rcomb[:,:,ic],u0pl)
                            bcom = matmul(tcomb[:,:,ic],utmi)
                            acom += bcom
                            umi = acom + jcomb[:,:,ic]
                            for imu in range(iemm, iemm+2): 
                                yx[ico] = umi[imu,0]
                                ico += 1
                                u0pl[imu0] = 0.0 
                    else:
                        for imu0 in range(isol, isol+2): 
                            
                            u0pl[imu0] = solar1[widx] / (2.0 * np.pi * wtmu[imu0])
                            if lowbc == 0: #We are at the very bottom of the atmosphere+surface system
                                acom = np.ascontiguousarray(tcomb[:,:,ic]) @ u0pl
                                bcom = np.ascontiguousarray(rcomb[:,:,ic]) @ utmi
                                acom += bcom
                                upl = acom + jcomb[:,:,ic]
                            else: #We are not at the very bottom, so we need to compute the internal radiation field
                                upl = idown(rcomb[:,:,ic],tcomb[:,:,ic],jcomb[:,:,ic],rs[:,:,ic],ts[:,:,ic],js[:,:,ic],u0pl,utmi,e)
                            
                            for imu in range(iemm, iemm+2): 
                                yx[ico] = upl[imu,0]
                                ico += 1
                                u0pl[imu0] = 0.0 

                    drad = ((1-t)*(1-u)*yx[0] + t*(1-u)*yx[1] + t*u*yx[3] + (1-t)*u*yx[2]) * np.cos(ic*aphi * np.pi / 180.0)

                    if ic > 0:
                        drad *= 2

                    rad[ipath,ig,widx] += drad
                    conv = np.abs(drad / rad[ipath,ig,widx])
                    
                    # If sure about convergence, exit loop over fourier components
                    if conv < defconv and conv1:
                        break

                    if conv < defconv:
                        conv1 = True
                    else:
                        conv1 = False
    
    return rad






#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
#
# archNEMESIS - Python implementation of the NEMESIS radiative transfer and retrieval code
# Surface_0.py - Object to represent the surface properties of a planet.
#
# Copyright (C) 2025 Juan Alday, Joseph Penn, Patrick Irwin,
# Jack Dobinson, Jon Mason, Jingxuan Yang
#
# This file is part of archNEMESIS.
#
# archNEMESIS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations #  for 3.9 compatability
from scipy.interpolate import interp1d
import numpy as np
import matplotlib.pyplot as plt
from numba import jit#, njit

from archnemesis.enums import WaveUnit, LowerBoundaryCondition
from archnemesis.helpers import h5py_helper

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.DEBUG)

#!/usr/local/bin/python3
# -*- coding: utf-8 -*-

###############################################################################################

"""
Created on Tue Jul 22 17:27:12 2021

@author: juanalday

[JD] We might want to rework this class to be a bit more general e.g. "LowerBoundary" so
     that it more accurately specifies what it does. That way it makes a bit more sense to use
     it for gas giants. Really, we should make an interface 'LowerBoundaryCondition' and then
     have subclasses for each of the different types of lower boundary conditions.
"""

class Surface_0:

    def __init__(self, GASGIANT=False, ISPACE=WaveUnit.Wavenumber_cm, LOWBC=LowerBoundaryCondition.LAMBERTIAN, GALB=-1.0, NEM=2, NLOCATIONS=1):

        """
        Inputs
        ------
        @param GASGIANT: log,
            Flag indicating whether the planet has surface (GASGIANT=False) or not (GASGIANT=True)

        @param ISPACE: int
            Spectral units
                0 :: Wavenumber (cm-1)
                1 :: Wavelength (um)

        @param LOWBC: int,
            Flag indicating the lower boundary condition.
                0 :: Thermal emission only (i.e. no reflection)
                1 :: Lambertian surface
                2 :: Hapke surface
 
        @param GALB: int,
            Ground albedo
            
        @param NEM: int,
            Number of spectral points defining the emissivity of the surface   

        @param NLOCATIONS: int,
            Number of surface points (i.e. different latitudes/longitudes with different properties)
        
        Attributes
        ----------

        @attribute LATITUDE: real or 1D array (depending on number of locations)
            Latitude of each location (degree)

        @attribute LONGITUDE: real or 1D array (depending on number of locations)
            Longitude of each location (degree)

        @attribute TSURF: real or 1D array (depending on number of locations)
            Surface temperature (K)

        Attributes for Thermal emission from surface:
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        @attribute VEM: 1D array
            Wavelengths at which the emissivity and other surface parameters are defined
            Assumed to be equal for all locations

        @attribute EMISSIVITY: 1D array or 2D array (depending on number of locations)
            Surface emissitivity 


        Attributes for Lambertian surface:
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        None. The albedo of a lambertian surface is given in the .set file by GALB, 
        and if GALB<0, it is calculated as 1.0 - EMISSIVITY.


        Attributes for Hapke surface:
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        The method implemented here is derived in Hapke (2012) - Theory of Reflectance and Emittance Spectroscopy
        In particular, it is derived in chapter 12.3.1 of that book. We also assume that the scattering phase function
        of the surface is given by a Double Henyey-Greenstein function. 

        @attribute SGLALB : 1D array or 2D array (depending on number of locations)
            Single scattering albedo w

        @attribute K : 1D array or 2D array (depending on number of locations)
            Porosity coefficient 

        @attribute BS0 : 1D array or 2D array (depending on number of locations)
            Amplitude of opposition effect (0 to 1)    

        @attribute hs : 1D array or 2D array (depending on number of locations)
            Width of the opposition surge

        @attribute BC0 : 1D array or 2D array (depending on number of locations)
            Amplitude of the coherent backscatter opposition effect (0 to 1)

        @attribute hc : 1D array or 2D array (depending on number of locations)
            Width of the backscatter function

        @attribute ROUGHNESS : 1D array or 2D array (depending on number of locations)
            Roughness mean slope angle (degrees)

        @attribute G1 : 1D array or 2D array (depending on number of locations)
            Asymmetry factor of the first Henyey-Greenstein function defining the phase function

        @attribute G2 : 1D array or 2D array (depending on number of locations)
            Asymmetry factor of the second Henyey-Greenstein function defining the phase function
        
        @attribute F: 1D array or 2D array (depending on number of locations)
            Parameter defining the relative contribution of G1 and G2 of the double Henyey-Greenstein phase function


        Attributes for Oren-Nayar surface reflection:
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        The method implemented here is described in Oren & Nayar (1998). This method is a generalisation of 
        the Lambertian reflection model to account for the roughness of the surface. In the case that the
        roughness is zero, this model converges to a Lambertian surface.

        @attribute ALBEDO : 1D array or 2D array (depending on number of locations)
            Lambert albedo
            
        @attribute ROUGHNESS : 1D array or 2D array (depending on number of locations)
            Roughness mean slope angle (degrees)

        Methods
        -------
        Surface_0.assess()

        Surface_0.write_hdf5()
        Surface_0.read_hdf5()

        Surface_0.read_sur()
        Surface_0.read_hap()

        Surface_0.edit_EMISSIVITY()
        Surface_0.edit_SGLALB()
        Surface_0.edit_BS0()
        Surface_0.edit_hs()
        Surface_0.edit_BC0()
        Surface_0.edit_hc()
        Surface_0.edit_K()
        Surface_0.edit_ROUGHNESS()
        Surface_0.edit_G1()
        Surface_0.edit_G2()
        Surface_0.edit_F()
        
        Surface_0.select_location()
        """

        #Input parameters
        self.NLOCATIONS = NLOCATIONS
        self.GASGIANT = GASGIANT
        #self.ISPACE = WaveUnit(ISPACE) if not isinstance(ISPACE, WaveUnit) else ISPACE
        #self.LOWBC = LowerBoundaryCondition(LOWBC) if not isinstance(LOWBC, LowerBoundaryCondition) else LOWBC
        self.GALB = GALB
        self.NEM = NEM

        # Input the following profiles using the edit_ methods.
        self.LATITUDE = None   #float or (NLOCATIONS) 
        self.LONGITUDE = None  #float or (NLOCATIONS) 
        self.TSURF = None      #float or (NLOCATIONS) 
        self.VEM = None #(NEM) or (NEM)
        self.EMISSIVITY = None #(NEM) or (NEM,NLOCATIONS)

        #Hapke parameters
        self.SGLALB = None #(NEM) or (NEM,NLOCATIONS)
        self.BS0 = None #(NEM) or (NEM,NLOCATIONS)
        self.hs = None #(NEM) or (NEM,NLOCATIONS)
        self.BC0 = None #(NEM) or (NEM,NLOCATIONS)
        self.hc = None #(NEM) or (NEM,NLOCATIONS)
        self.K = None #(NEM) or (NEM,NLOCATIONS)
        self.ROUGHNESS = None #(NEM) or (NEM,NLOCATIONS)
        self.G1 = None #(NEM) or (NEM,NLOCATIONS)
        self.G2 = None #(NEM) or (NEM,NLOCATIONS)
        self.F = None #(NEM) or (NEM,NLOCATIONS)

        # private attributes
        self._ispace = None
        self._lowbc = None
        
        # set property values
        self.ISPACE = ISPACE if ISPACE is not None else WaveUnit.Wavenumber_cm
        self.LOWBC = LOWBC if LOWBC is not None else LowerBoundaryCondition.THERMAL
    
    
    @property
    def ISPACE(self) -> WaveUnit:
        return self._ispace
    
    @ISPACE.setter
    def ISPACE(self, value):
        self._ispace = WaveUnit(value)
    
    @property
    def LOWBC(self) -> LowerBoundaryCondition:
        return self._lowbc
    
    @LOWBC.setter
    def LOWBC(self, value):
        self._lowbc = LowerBoundaryCondition(value)
        


    def assess(self):
        """
        Assess whether the different variables have the correct dimensions and types
        """

        if self.GASGIANT==False:

            #Checking some common parameters to all cases
            assert isinstance(self.LOWBC, (int, np.integer, LowerBoundaryCondition)), 'LOWBC must be int or LowerBoundaryCondition'
            assert self.LOWBC in LowerBoundaryCondition, f'LOWBC must be one of {tuple(LowerBoundaryCondition)}'
            
            assert len(self.VEM) == self.NEM , 'VEM must have size (NEM)'
            
            assert isinstance(self.ISPACE, (int, np.integer, WaveUnit)), 'ISPACE must be int or WaveUnit'
            assert self.ISPACE in WaveUnit, f'ISPACE must be one of {tuple(WaveUnit)}'

            #Determining sizes based on the number of surface locations
            if self.NLOCATIONS<=0:
                raise ValueError('error :: NLOCATIONS must be greater than 0')

            elif self.NLOCATIONS==1:

                assert np.issubdtype(type(self.TSURF), np.float64) == True , \
                    'TSURF must be float'
                assert np.issubdtype(type(self.LATITUDE), np.float64) == True , \
                    'LATITUDE must be float'
                assert abs(self.LATITUDE) < 90.0 , \
                    'LATITUDE must be within -90 to 90 degrees'
                assert np.issubdtype(type(self.LONGITUDE), np.float64) == True , \
                    'LONGITUDE must be float'

                assert len(self.EMISSIVITY) == self.NEM , \
                    'EMISSIVITY must have size (NEM)'

                #Special case for Hapke reflection
                if self.LOWBC==LowerBoundaryCondition.HAPKE:
                    assert len(self.SGLALB) == self.NEM , \
                        'SGLALB must have size (NEM)'
                    assert len(self.ROUGHNESS) == self.NEM , \
                        'ROUGHNESS must have size (NEM)'
                    assert len(self.BS0) == self.NEM , \
                        'BS0 must have size (NEM)'
                    assert len(self.hs) == self.NEM , \
                        'hs must have size (NEM)'
                    assert len(self.BC0) == self.NEM , \
                        'BC0 must have size (NEM)'
                    assert len(self.hc) == self.NEM , \
                        'hc must have size (NEM)'
                    assert len(self.K) == self.NEM , \
                        'K must have size (NEM)'
                    assert len(self.G1) == self.NEM , \
                        'G1 must have size (NEM)'
                    assert len(self.G2) == self.NEM , \
                        'G2 must have size (NEM)'
                    assert len(self.F) == self.NEM , \
                        'F must have size (NEM)'
                        
                elif self.LOWBC==3:
                    assert len(self.ALBEDO) == self.NEM , \
                        'ALBEDO must have size (NEM)'
                    assert len(self.ROUGHNESS) == self.NEM , \
                        'ROUGHNESS must have size (NEM)'
                        
            else:
                assert len(self.TSURF) == self.NLOCATIONS , \
                    'TSURF must have size (NLOCATIONS)'
                assert len(self.LATITUDE) == self.NLOCATIONS , \
                    'LATITUDE must have size (NLOCATIONS)'
                assert len(self.LONGITUDE) == self.NLOCATIONS , \
                    'LONGITUDE must have size (NLOCATIONS)'
                
                assert self.EMISSIVITY.shape == (self.NEM,self.NLOCATIONS) , \
                    'EMISSIVITY must have size (NEM,NLOCATIONS)'
                
                #Special case for Hapke reflection
                if self.LOWBC==LowerBoundaryCondition.HAPKE:
                    assert self.SGLALB.shape == (self.NEM,self.NLOCATIONS) , \
                        'SGLALB must have size (NEM,NLOCATIONS)'
                    assert self.BS0.shape == (self.NEM,self.NLOCATIONS) , \
                        'BS0 must have size (NEM,NLOCATIONS)'
                    assert self.hs.shape == (self.NEM,self.NLOCATIONS) , \
                        'hs must have size (NEM,NLOCATIONS)'
                    assert self.BC0.shape == (self.NEM,self.NLOCATIONS) , \
                        'BC0 must have size (NEM,NLOCATIONS)'
                    assert self.hc.shape == (self.NEM,self.NLOCATIONS) , \
                        'hc must have size (NEM,NLOCATIONS)'
                    assert self.K.shape == (self.NEM,self.NLOCATIONS) , \
                        'K must have size (NEM,NLOCATIONS)'
                    assert self.ROUGHNESS.shape == (self.NEM,self.NLOCATIONS) , \
                        'ROUGHNESS must have size (NEM,NLOCATIONS)'
                    assert self.G1.shape == (self.NEM,self.NLOCATIONS) , \
                        'G1 must have size (NEM,NLOCATIONS)'
                    assert self.G2.shape == (self.NEM,self.NLOCATIONS) , \
                        'G2 must have size (NEM,NLOCATIONS)'
                    assert self.F.shape == (self.NEM,self.NLOCATIONS) , \
                        'F must have size (NEM,NLOCATIONS)'
                        
                #Special case for Oren-Nayar reflection
                elif self.LOWBC==3:
                    assert self.ALBEDO.shape == (self.NEM,self.NLOCATIONS) , \
                        'ALBEDO must have size (NEM,NLOCATIONS)'
                    assert self.ROUGHNESS.shape == (self.NEM,self.NLOCATIONS) , \
                        'ROUGHNESS must have size (NEM,NLOCATIONS)'

        else:
            assert self.TSURF == 0.0, \
                "if GASGIANT=True then no emission from surface"
            assert self.GALB == 0.0, \
                "if GASGIANT=True then surface absorbs everything"
            assert self.LOWBC == LowerBoundaryCondition.THERMAL , \
                'If GASGIANT=True then LOWBC=0 (i.e. No reflection)'

    def write_hdf5(self,runname):
        """
        Write the surface properties into an HDF5 file
        """

        import h5py

        #Assessing that all the parameters have the correct type and dimension
        self.assess()

        with h5py.File(runname+'.h5','a') as f:
            #Checking if Atmosphere already exists
            if ('/Surface' in f)==True:
                del f['Surface']   #Deleting the Atmosphere information that was previously written in the file

            if self.GASGIANT==False:

                grp = f.create_group("Surface")

                #Writing the lower boundary condition
                dset = h5py_helper.store_data(grp, 'LOWBC', int(self.LOWBC))
                dset.attrs['title'] = "Lower Boundary Condition"
                if self.LOWBC==LowerBoundaryCondition.THERMAL:
                    dset.attrs['type'] = 'Isotropic thermal emission (no reflection)'
                elif self.LOWBC==LowerBoundaryCondition.LAMBERTIAN:
                    dset.attrs['type'] = 'Isotropic thermal emission and Lambert reflection'
                elif self.LOWBC==LowerBoundaryCondition.HAPKE:
                    dset.attrs['type'] = 'Isotropic thermal emission and Hapke reflection'
                elif self.LOWBC==LowerBoundaryCondition.OREN_NAYAR:
                    dset.attrs['type'] = 'Isotropic thermal emission and Oren-Nayar reflection'

                #Writing the spectral units
                dset = h5py_helper.store_data(grp, 'ISPACE', int(self.ISPACE))
                dset.attrs['title'] = "Spectral units"
                if self.ISPACE==WaveUnit.Wavenumber_cm:
                    dset.attrs['units'] = 'Wavenumber / cm-1'
                elif self.ISPACE==WaveUnit.Wavelength_um:
                    dset.attrs['units'] = 'Wavelength / um'

                #Writing the spectral array
                dset = h5py_helper.store_data(grp, 'VEM', self.VEM)
                dset.attrs['title'] = "Spectral array"
                if self.ISPACE==WaveUnit.Wavenumber_cm:
                    dset.attrs['units'] = 'Wavenumber / cm-1'
                elif self.ISPACE==WaveUnit.Wavelength_um:
                    dset.attrs['units'] = 'Wavelength / um'

                #Writing the number of locations
                dset = h5py_helper.store_data(grp, 'NLOCATIONS', self.NLOCATIONS)
                dset.attrs['title'] = "Number of surface locations"

                #Writing the co-ordinates of the locations
                dset = h5py_helper.store_data(grp, 'LATITUDE', self.LATITUDE)
                dset.attrs['title'] = "Latitude of the surface locations"
                dset.attrs['units'] = 'degrees'

                dset = h5py_helper.store_data(grp, 'LONGITUDE', self.LONGITUDE)
                dset.attrs['title'] = "Longitude of the surface locations"
                dset.attrs['units'] = 'degrees'

                #Writing the surface temperature
                dset = h5py_helper.store_data(grp, 'TSURF', self.TSURF)
                dset.attrs['title'] = "Surface Temperature"
                dset.attrs['units'] = 'K'

                #Writing the surface albedo
                dset = h5py_helper.store_data(grp, 'GALB', self.GALB)
                dset.attrs['title'] = "Ground albedo"
                if self.GALB<0.0:
                    dset.attrs['type'] = 'Surface albedo calculated as (1 - EMISSIVITY)'

                #Writing the emissivity
                dset = h5py_helper.store_data(grp, 'EMISSIVITY', self.EMISSIVITY)
                dset.attrs['title'] = "Surface emissivity"
                dset.attrs['units'] = ''

                #Writing Hapke parameters if they are required
                if self.LOWBC==LowerBoundaryCondition.HAPKE:

                    dset = h5py_helper.store_data(grp, 'SGLALB', self.SGLALB)
                    dset.attrs['title'] = "Single scattering albedo"
                    dset.attrs['units'] = ''

                    dset = h5py_helper.store_data(grp, 'K', self.K)
                    dset.attrs['title'] = "Porosity coefficient"
                    dset.attrs['units'] = ''

                    dset = h5py_helper.store_data(grp, 'BS0', self.BS0)
                    dset.attrs['title'] = "Amplitude of the opposition effect"
                    dset.attrs['units'] = ''

                    dset = h5py_helper.store_data(grp, 'hs', self.hs)
                    dset.attrs['title'] = "Width of the opposition surge"
                    dset.attrs['units'] = ''

                    dset = h5py_helper.store_data(grp, 'BC0', self.BC0)
                    dset.attrs['title'] = "Amplitude of the coherent backscatter opposition effect"
                    dset.attrs['units'] = ''

                    dset = h5py_helper.store_data(grp, 'hc', self.hc)
                    dset.attrs['title'] = "Width of the backscatter function"
                    dset.attrs['units'] = ''

                    dset = h5py_helper.store_data(grp, 'ROUGHNESS', self.ROUGHNESS)
                    dset.attrs['title'] = "Roughness mean slope angle"
                    dset.attrs['units'] = 'degrees'

                    dset = h5py_helper.store_data(grp, 'G1', self.G1)
                    dset.attrs['title'] = "Asymmetry factor of the first Henyey-Greenstein function defining the phase function"
                    dset.attrs['units'] = ''

                    dset = h5py_helper.store_data(grp, 'G2', self.G2)
                    dset.attrs['title'] = "Asymmetry factor of the second Henyey-Greenstein function defining the phase function"
                    dset.attrs['units'] = ''

                    dset = h5py_helper.store_data(grp, 'F', self.F)
                    dset.attrs['title'] = "Parameter defining the relative contribution of G1 and G2 of the double Henyey-Greenstein phase function"
                    dset.attrs['units'] = ''

                elif self.LOWBC==LowerBoundaryCondition.OREN_NAYAR:

                    dset = h5py_helper.store_data(grp, 'ALBEDO', self.ALBEDO)
                    dset.attrs['title'] = "Surface albedo"
                    dset.attrs['units'] = ''
                    
                    dset = h5py_helper.store_data(grp, 'ROUGHNESS', self.ROUGHNESS)
                    dset.attrs['title'] = "Roughness mean slope angle"
                    dset.attrs['units'] = 'degrees'


    def read_hdf5(self,runname):
        """
        Read the surface properties from an HDF5 file
        """

        import h5py

        with h5py.File(runname+'.h5','r') as f:

            #Checking if Surface exists
            e = "/Surface" in f
            if e==False:
                # If we have "no surface", planet is a gasgiant
                # so the surface is always perfectly isothermal
                # and absorbing.
                self.GASGIANT = True
                self.LOWBC = LowerBoundaryCondition.THERMAL
                self.TSURF = 0.0
                self.GALB = 0.0
            else:
                self.GASGIANT = False
                self.ISPACE = h5py_helper.retrieve_data(f, 'Surface/ISPACE', lambda x:  WaveUnit(np.int32(x)))
                self.LOWBC = h5py_helper.retrieve_data(f, 'Surface/LOWBC', lambda x:  LowerBoundaryCondition(np.int32(x)))
                self.NLOCATIONS = h5py_helper.retrieve_data(f, 'Surface/NLOCATIONS', np.int32)

                self.VEM = h5py_helper.retrieve_data(f, 'Surface/VEM', np.array)
                self.NEM = len(self.VEM)
                if self.NLOCATIONS==1:
                    self.TSURF = h5py_helper.retrieve_data(f, 'Surface/TSURF', np.float64)
                    self.LATITUDE = h5py_helper.retrieve_data(f, 'Surface/LATITUDE', np.float64)
                    self.LONGITUDE = h5py_helper.retrieve_data(f, 'Surface/LONGITUDE', np.float64)
                else:
                    self.TSURF = h5py_helper.retrieve_data(f, 'Surface/TSURF', np.array)
                    self.LATITUDE = h5py_helper.retrieve_data(f, 'Surface/LATITUDE', np.array)
                    self.LONGITUDE = h5py_helper.retrieve_data(f, 'Surface/LONGITUDE', np.array)

                self.EMISSIVITY = h5py_helper.retrieve_data(f, 'Surface/EMISSIVITY', np.array)

                if self.LOWBC==LowerBoundaryCondition.LAMBERTIAN:
                    self.GALB = h5py_helper.retrieve_data(f, 'Surface/GALB', np.array)

                if self.LOWBC==LowerBoundaryCondition.HAPKE:
                    self.SGLALB = h5py_helper.retrieve_data(f, 'Surface/SGLALB', np.array)
                    self.BS0 = h5py_helper.retrieve_data(f, 'Surface/BS0', np.array)
                    self.hs = h5py_helper.retrieve_data(f, 'Surface/hs', np.array)
                    self.BC0 = h5py_helper.retrieve_data(f, 'Surface/BC0', np.array)
                    self.hc = h5py_helper.retrieve_data(f, 'Surface/hc', np.array)
                    self.K = h5py_helper.retrieve_data(f, 'Surface/K', np.array)
                    self.ROUGHNESS = h5py_helper.retrieve_data(f, 'Surface/ROUGHNESS', np.array)
                    self.G1 = h5py_helper.retrieve_data(f, 'Surface/G1', np.array)
                    self.G2 = h5py_helper.retrieve_data(f, 'Surface/G2', np.array)
                    self.F = h5py_helper.retrieve_data(f, 'Surface/F', np.array)
                    
                if self.LOWBC==LowerBoundaryCondition.OREN_NAYAR:
                    self.ALBEDO = h5py_helper.retrieve_data(f, 'Surface/ALBEDO', np.array)
                    self.ROUGHNESS = h5py_helper.retrieve_data(f, 'Surface/ROUGHNESS', np.array)

        self.assess()


    def edit_EMISSIVITY(self, EMISSIVITY_array):
        """
        Edit the surface emissivity at each of the lat/lon points
        @param EMISSIVITY_array: 3D array
            Array defining the surface emissivity at each of the points
        """
        EMISSIVITY_array = np.array(EMISSIVITY_array)
        if self.NLOCATIONS==1:
            assert len(EMISSIVITY_array) == self.NEM , \
                'EMISSIVITY should have NEM elements'
        else:
            assert EMISSIVITY_array.shape == (self.NEM,self.NLOCATIONS) , \
                'EMISSIVITY should have (NEM,NLOCATIONS) elements'
        self.EMISSIVITY = EMISSIVITY_array 


    def edit_SGLALB(self, array):
        """
        Edit the single scattering albedo at each of the lat/lon points
        @param array: 1D or 2D array
        """
        array = np.array(array)
        if self.NLOCATIONS==1:
            assert len(array) == self.NEM , \
                'SGLALB should have NEM elements'
        else:
            assert array.shape == (self.NEM,self.NLOCATIONS) , \
                'SGLALB should have (NEM,NLOCATIONS) elements'
        self.SGLALB = array 

    def edit_ROUGHNESS(self, array):
        """
        Edit the roughness mean slope angle at each of the lat/lon points
        @param array: 1D or 2D array
        """
        array = np.array(array)
        if self.NLOCATIONS==1:
            assert len(array) == self.NEM , \
                'ROUGHNESS should have NEM elements'
        else:
            assert array.shape == (self.NEM,self.NLOCATIONS) , \
                'ROUGHNESS should have (NEM,NLOCATIONS) elements'
        self.ROUGHNESS = array 

    def edit_BS0(self, array):
        """
        Edit the amplitude of the opposition effect at each of the lat/lon points
        @param array: 1D or 2D array
        """
        array = np.array(array)
        if self.NLOCATIONS==1:
            assert len(array) == self.NEM , \
                'BS0 should have NEM elements'
        else:
            assert array.shape == (self.NEM,self.NLOCATIONS) , \
                'BS0 should have (NEM,NLOCATIONS) elements'
        self.BS0 = array 

    def edit_hs(self, array):
        """
        Edit the width of the opposition effect at each of the lat/lon points
        @param array: 1D or 2D array
        """
        array = np.array(array)
        if self.NLOCATIONS==1:
            assert len(array) == self.NEM , \
                'hs should have NEM elements'
        else:
            assert array.shape == (self.NEM,self.NLOCATIONS) , \
                'hs should have (NEM,NLOCATIONS) elements'
        self.hs = array 

    def edit_BC0(self, array):
        """
        Edit the amplitude of the backscatter opposition effect at each of the lat/lon points
        @param array: 1D or 2D array
        """
        array = np.array(array)
        if self.NLOCATIONS==1:
            assert len(array) == self.NEM , \
                'BC0 should have NEM elements'
        else:
            assert array.shape == (self.NEM,self.NLOCATIONS) , \
                'BC0 should have (NEM,NLOCATIONS) elements'
        self.BC0 = array 

    def edit_hc(self, array):
        """
        Edit the width of the backscatter opposition effect at each of the lat/lon points
        @param array: 1D or 2D array
        """
        array = np.array(array)
        if self.NLOCATIONS==1:
            assert len(array) == self.NEM , \
                'hc should have NEM elements'
        else:
            assert array.shape == (self.NEM,self.NLOCATIONS) , \
                'hc should have (NEM,NLOCATIONS) elements'
        self.hc = array 

    def edit_K(self, array):
        """
        Edit the porosity coefficient at each of the lat/lon points
        @param array: 1D or 2D array
        """
        array = np.array(array)
        if self.NLOCATIONS==1:
            assert len(array) == self.NEM , \
                'K should have NEM elements'
        else:
            assert array.shape == (self.NEM,self.NLOCATIONS) , \
                'K should have (NEM,NLOCATIONS) elements'
        self.K = array 

    def edit_G1(self, array):
        """
        Edit the first assymmetry parameter at each of the lat/lon points
        @param array: 1D or 2D array
        """
        array = np.array(array)
        if self.NLOCATIONS==1:
            assert len(array) == self.NEM , \
                'G1 should have NEM elements'
        else:
            assert array.shape == (self.NEM,self.NLOCATIONS) , \
                'G1 should have (NEM,NLOCATIONS) elements'
        self.G1 = array 

    def edit_G2(self, array):
        """
        Edit the second assymmetry parameter at each of the lat/lon points
        @param array: 1D or 2D array
        """
        array = np.array(array)
        if self.NLOCATIONS==1:
            assert len(array) == self.NEM , \
                'G2 should have NEM elements'
        else:
            assert array.shape == (self.NEM,self.NLOCATIONS) , \
                'G2 should have (NEM,NLOCATIONS) elements'
        self.G2 = array 

    def edit_F(self, array):
        """
        Edit the contribution from each H-G function at each of the lat/lon points
        @param array: 1D or 2D array
        """
        array = np.array(array)
        if self.NLOCATIONS==1:
            assert len(array) == self.NEM , \
                'F should have NEM elements'
        else:
            assert array.shape == (self.NEM,self.NLOCATIONS) , \
                'F should have (NEM,NLOCATIONS) elements'
        self.F = array 

    def calc_phase_angle(self,EMISS_ANG,SOL_ANG,AZI_ANG):
        """
        Calculate the phase angle based on the emission, incident and azimuth angles

        Inputs
        ------
        @param EMISS_ANG: 1D array or scalar
            Emission angle (deg)
        @param SOL_ANG: 1D array or scalar
            Solar zenith or incident angle (deg)
        @param AZI_ANG: 1D array or scalar
            Azimuth angle (deg)

        Outputs
        -------
        @param PHASE_ANG: 1D array or scalar
            Phase angle (deg)
        """

        #First of all let's calculate the scattering phase angle
        mu = np.cos(EMISS_ANG/180.*np.pi)   #Cosine of the reflection angle
        mu0 = np.cos(SOL_ANG/180.*np.pi)    #Coside of the incidence angle

        cg = mu * mu0 + np.sqrt(1. - mu**2.) * np.sqrt(1.-mu0**2.) * np.cos(AZI_ANG/180.*np.pi)
        iin = np.where(cg>1.0)
        cg[iin] = 1.0 
        g = np.arccos(cg)/np.pi*180.   #Scattering phase angle (degrees) (NTHETA)

        return g


    def select_location(self,iLOCATION):
        """
        Subroutine to select only one geometry from the Atmosphere class (and remove all the others)

        Inputs
        ------
        @param iLOCATION: int
            Index of the location to be selected

        Outputs
        -------
        Updated Surface class
        """
        
        if iLOCATION>self.NLOCATIONS-1:
            raise ValueError('error in select_location :: iLOCATION must be between 0 and NLOCATIONS-1',[0,self.NLOCATIONS-1])

        self.NLOCATIONS = 1
        
        self.LATITUDE = self.LATITUDE[iLOCATION]
        self.LONGITUDE = self.LONGITUDE[iLOCATION]
        self.TSURF = self.TSURF[iLOCATION]
        
        self.edit_EMISSIVITY(self.EMISSIVITY[:,iLOCATION])
        
        if self.LOWBC==LowerBoundaryCondition.HAPKE: #Hapke case
            
            self.edit_SGLALB(self.SGLALB[:,iLOCATION])
            self.edit_BS0(self.BS0[:,iLOCATION])
            self.edit_hs(self.hs[:,iLOCATION])
            self.edit_BC0(self.BC0[:,iLOCATION])
            self.edit_hc(self.hc[:,iLOCATION])
            self.edit_K(self.K[:,iLOCATION])
            self.edit_ROUGHNESS(self.ROUGHNESS[:,iLOCATION])
            self.edit_G1(self.G1[:,iLOCATION])
            self.edit_G2(self.G2[:,iLOCATION])
            self.edit_F(self.F[:,iLOCATION])
            
        if self.LOWBC==3: #Oren-Nayar case
            
            self.edit_ALBEDO(self.ALBEDO[:,iLOCATION])
            self.edit_ROUGHNESS(self.ROUGHNESS[:,iLOCATION])
            
        #Checking that everything went well
        self.assess()
            
        


    ##################################################################################################################
    ##################################################################################################################
    #                                                   EMISSIVITY
    ##################################################################################################################
    ##################################################################################################################

    def read_sur(self, runname):
        """
        Read the surface emissivity from the .sur file
        @param runname: str
            Name of the Nemesis run
        """
        
        #Opening file
        f = open(runname+'.sur','r')
        nem = int(np.fromfile(f,sep=' ',count=1,dtype='int'))
    
        vem = np.zeros([nem])
        emissivity = np.zeros([nem])
        for i in range(nem):
            tmp = np.fromfile(f,sep=' ',count=2,dtype='float')
            vem[i] = tmp[0]
            emissivity[i] = tmp[1]

        self.NEM = nem
        self.VEM = vem
        self.EMISSIVITY = emissivity

    ##################################################################################################################

    def write_sur(self, runname):
        """
        Write the surface emissivity into the .sur file
        @param runname: str
            Name of the Nemesis run
        """
        
        #Opening file
        f = open(runname+'.sur','w')
        f.write('%i \n' % (self.NEM))
        for i in range(self.NEM):
            f.write('%7.4e \t %7.4e \n' % (self.VEM[i],self.EMISSIVITY[i]))
        f.close()

    ##################################################################################################################


    def calc_radground(self,ISPACE,WAVE=None):
        """
        Calculate the thermal emission from the surface

        @param ISPACE: int
            Units of wavelength array
            (0) Wavenumber in cm-1 (1) Wavelength in microns

        @param WAVE: 1D array
            Wavelength or wavenumber array
        """

        if WAVE is None:
            WAVE = self.VEM

        bbsurf = planck(ISPACE,WAVE,self.TSURF)

        # bbsurf has shape (self.NEM, self.NLOCATIONS)
        # so self.EMISSIVITY must have that shape as well if it is 2D
        # Therefore, spectral axis is 0th axis.
        f = interp1d(self.VEM,self.EMISSIVITY, axis=0)
        emissivity = f(WAVE)

        radground = bbsurf * emissivity

        return radground


    ##################################################################################################################
    ##################################################################################################################
    #                            BIDIRECTIONAL REFLECTANCE DISTRIBUTION FUNCTION (BRDF)
    ##################################################################################################################
    ##################################################################################################################

    def calc_reflectance(self,WAVE,SOL_ANG,EMISS_ANG,AZI_ANG,E0=None):
        """
        Calculate the reflected radiance from the surface based on the type of surface defined
        
        Inputs
        ------
        @param WAVE(NWAVE): 1D array
            Wavelengths (microns) or Wavenumbers (cm-1)
        @param SOL_ANG(NTHETA): 1D array
            Solar zenith angle (degrees)
        @param EMISS_ANG(NTHETA): 1D array
            Emission angle (degrees)
        @param AZI_ANG(NTHETA): 1D array    
            Azimuth angle (degrees)
            
        Optional inputs
        ---------------
        @param E0(NWAVE): 1D array
            Solar spectral irradiance (W/m2/cm-1) or (W/m2/um)
            
        Outputs
        -------
        @param R(NWAVE,NTHETA): 2D array
            Reflected radiance (W/m2/cm-1/sr or W/m2/um/sr) 
        """
        
        NWAVE = len(WAVE)
        #NTHETA = len(EMISS_ANG)
        
        if E0 is None:
            E0 = np.ones(NWAVE)
        else:
            if len(E0) != NWAVE:
                raise ValueError('error in calc_reflectance :: E0 must have the same size as WAVE')
            
        BRDF = self.calc_BRDF(WAVE,SOL_ANG,EMISS_ANG,AZI_ANG)  #Bidirectional reflectance distribution function
        
        R = BRDF * np.cos(SOL_ANG/180.*np.pi)
        R *= E0[:,None]
        
        # Set R to 0 where SOL_ANG > 90
        R[:, SOL_ANG > 90] = 0
        
        return R


    def calc_BRDF(self,WAVE,SOL_ANG,EMISS_ANG,AZI_ANG):
        """
        Calculate the BRDF of the surface based on the type of surface defined
        
        Inputs
        ------
        @param WAVE(NWAVE): 1D array
            Wavelengths (microns) or Wavenumbers (cm-1)
        @param SOL_ANG(NTHETA): 1D array
            Solar zenith angle (degrees)
        @param EMISS_ANG(NTHETA): 1D array
            Emission angle (degrees)
        @param AZI_ANG(NTHETA): 1D array    
            Azimuth angle (degrees)
            
        Outputs
        -------
        @param BRDF(NWAVE,NTHETA): 2D array
            Bidirectional reflectance distribution function (BRDF)
        """
        
        NWAVE = len(WAVE)
        NTHETA = len(EMISS_ANG)
        
        BRDF = np.zeros((NWAVE,NTHETA))
        if self.LOWBC==LowerBoundaryCondition.LAMBERTIAN: #Lambertian reflection
            
            ALBEDO = self.calc_albedo()                 #Calculating albedo based on flags in class
            galbx = np.interp(WAVE,self.VEM,ALBEDO)  #Interpolating albedo to desired spectral array
            
            # Broadcasting to avoid loop
            BRDF[:,:] = (galbx[:, None] / np.pi)  # Shape (NWAVE, 1) broadcasted to (NWAVE, NTHETA)
            
        elif self.LOWBC==LowerBoundaryCondition.HAPKE: #Hapke reflection
            
            #Interpolating Hapke parameters to the desired spectral array
            SGLALB = np.interp(WAVE,self.VEM,self.SGLALB)
            K = np.interp(WAVE,self.VEM,self.K)
            BS0 = np.interp(WAVE,self.VEM,self.BS0)
            hs = np.interp(WAVE,self.VEM,self.hs)
            BC0 = np.interp(WAVE,self.VEM,self.BC0)
            hc = np.interp(WAVE,self.VEM,self.hc)
            ROUGHNESS = np.interp(WAVE,self.VEM,self.ROUGHNESS)
            G1 = np.interp(WAVE,self.VEM,self.G1)
            G2 = np.interp(WAVE,self.VEM,self.G2)
            F = np.interp(WAVE,self.VEM,self.F)
            
            #Calling the fortran module to calculate Hapke's BRDF
            BRDF[:,:] = calc_Hapke_BRDF(SGLALB,K,BS0,hs,BC0,hc,ROUGHNESS,G1,G2,F,\
                                    SOL_ANG,EMISS_ANG,AZI_ANG)
            
        elif self.LOWBC == 3: #Oren & Nayar reflection model
            
            #Interpolating Hapke parameters to the desired spectral array
            ALBEDO = np.interp(WAVE,self.VEM,self.ALBEDO)
            ROUGHNESS = np.interp(WAVE,self.VEM,self.ROUGHNESS)
            
            #Calling the fortran module to calculate Oren-Nayar's BRDF
            BRDF[:,:] = calc_OrenNayar_BRDF(ALBEDO,ROUGHNESS,SOL_ANG,EMISS_ANG,AZI_ANG)

        return BRDF


    ##################################################################################################################

    def calc_albedo(self):
        """
        Calculate the Lambert albedo of the surface based on the value on the class

        If GALB<0.0 then the Lambert albedo is calculated from the surface emissivity
        """

        if self.GALB>=0.0:
            ALBEDO = np.ones(self.NEM)*self.GALB
        else:
            ALBEDO = np.zeros(self.NEM)
            ALBEDO[:] = 1.0 - self.EMISSIVITY[:]

        return ALBEDO


    ##################################################################################################################
    ##################################################################################################################
    #                                         HAPKE BIDIRECTIONAL-REFLECTANCE
    ##################################################################################################################
    ##################################################################################################################    


    def read_hap(self, runname):
        """
        Read the Hapke parameters of the surface from the .hap file
        @param runname: str
            Name of the Nemesis run
        """
        
        #Opening file
        f = open(runname+'.hap','r')

        #Reading number of wavelengths
        nem = int(np.fromfile(f,sep=' ',count=1,dtype='int'))
    
        #Defining all fields
        vem = np.zeros(nem)
        sglalb = np.zeros(nem)
        k = np.zeros(nem)
        bso = np.zeros(nem)
        hs = np.zeros(nem)
        bco = np.zeros(nem)
        hc = np.zeros(nem)
        roughness = np.zeros(nem)
        g1 = np.zeros(nem)
        g2 = np.zeros(nem)
        fhg = np.zeros(nem)

        #Reading Hapke parameters
        for i in range(nem):
            tmp = np.fromfile(f,sep=' ',count=11,dtype='float')
            vem[i] = tmp[0]
            sglalb[i] = tmp[1]
            k[i] = tmp[2]
            bso[i] = tmp[3]
            hs[i] = tmp[4]
            bco[i] = tmp[5]
            hc[i] = tmp[6]
            roughness[i] = tmp[7]
            g1[i] = tmp[8]
            g2[i] = tmp[9]
            fhg[i] = tmp[10]

        f.close()

        #Storing parameters in the class
        self.NEM = nem
        self.VEM = vem
        self.SGLALB = sglalb
        self.K = k
        self.BS0 = bso
        self.hs = hs
        self.BC0 = bco
        self.hc = hc
        self.ROUGHNESS = roughness
        self.G1 = g1
        self.G2 = g2
        self.F = fhg

    ##################################################################################################################

    def write_hap(self,runname):
        """
        Read the Hapke parameters stored in the class into the .hap file
        @param runname: str
            Name of the Nemesis run
        """

        f = open(runname+'.hap','w')
        f.write('%i \n' % (self.NEM))
        for i in range(self.NEM):
            f.write('%7.4e \t %7.4e \t %7.4e \t %7.4e \t %7.4e \t %7.4e \t %7.4e \t %7.4e \t %7.4e \t %7.4e \t %7.4e \n' % \
                (self.VEM[i],self.SGLALB[i],self.K[i],self.BS0[i],self.hs[i],self.BC0[i],self.hc[i],self.ROUGHNESS[i],self.G1[i],self.G2[i],self.F[i]))
        f.close()


    ##################################################################################################################
    ##################################################################################################################
    #                                              PLOTTING FUNCTIONS
    ##################################################################################################################
    ################################################################################################################## 

    def plot_tsurf_map(self,subobs_lat=None,subobs_lon=None,cmap='viridis'):
        """
        Function to plot the surface temperature on a map 
        """

        from mpl_toolkits.basemap import Basemap
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        fig,ax1 = plt.subplots(1,1,figsize=(4,4))

        #Plotting the geometry
        if((subobs_lat is not None) & (subobs_lon is not None)):
            map = Basemap(projection='ortho', resolution=None,
                lat_0=subobs_lat, lon_0=subobs_lon)
        else:
            map = Basemap(projection='ortho', resolution=None,
                lat_0=np.mean(self.LATITUDE), lon_0=np.mean(self.LONGITUDE))
            
        map.drawparallels(np.linspace(-90, 90, 13)) # lats
        map.drawmeridians(np.linspace(-180, 180, 13)) # lons

        im = map.scatter(self.LONGITUDE,self.LATITUDE,latlon=True,c=self.TSURF,cmap=cmap)

        # create an axes on the right side of ax. The width of cax will be 5%
        # of ax and the padding between cax and ax will be fixed at 0.05 inch.
        divider = make_axes_locatable(ax1)
        cax = divider.append_axes("bottom", size="5%", pad=0.15)
        cbar2 = plt.colorbar(im,cax=cax,orientation='horizontal')
        cbar2.set_label('Surface Temperature (K)')

        ax1.grid()
        plt.tight_layout()
        plt.show()
        
    def plot_emissivity_map(self,subobs_lat=None,subobs_lon=None,cmap='viridis',iWAVE=0):
        """
        Function to plot the surface emissivity on a map 
        """

        from mpl_toolkits.basemap import Basemap
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        fig,ax1 = plt.subplots(1,1,figsize=(4,4))

        #Plotting the geometry
        if((subobs_lat is not None) & (subobs_lon is not None)):
            map = Basemap(projection='ortho', resolution=None,
                lat_0=subobs_lat, lon_0=subobs_lon)
        else:
            map = Basemap(projection='ortho', resolution=None,
                lat_0=np.mean(self.LATITUDE), lon_0=np.mean(self.LONGITUDE))
            
        map.drawparallels(np.linspace(-90, 90, 13)) # lats
        map.drawmeridians(np.linspace(-180, 180, 13)) # lons

        im = map.scatter(self.LONGITUDE,self.LATITUDE,latlon=True,c=self.EMISSIVITY[iWAVE,:],cmap=cmap)

        # create an axes on the right side of ax. The width of cax will be 5%
        # of ax and the padding between cax and ax will be fixed at 0.05 inch.
        divider = make_axes_locatable(ax1)
        cax = divider.append_axes("bottom", size="5%", pad=0.15)
        cbar2 = plt.colorbar(im,cax=cax,orientation='horizontal')
        cbar2.set_label('Surface Emissivity')

        ax1.grid()
        plt.tight_layout()
        plt.show()
        
        
    
        
        
################################################################################################################################
################################################################################################################################
#                                              OTHER FUNCTIONS OUTSIDE CLASS
################################################################################################################################
################################################################################################################################

#Conditional JIT compilation
# Define a flag to control JIT compilation
use_jit = True

def conditional_jit(nopython=True):
    def decorator(func):
        if use_jit:
            return jit(nopython=nopython)(func)
        else:
            return func
    return decorator


##################################################################################################################
#Thermal emission
##################################################################################################################

def planck(ispace,wave,temp):

    """
    FUNCTION NAME : planck()

    DESCRIPTION : Function to calculate the blackbody radiation given by the Planck function

    INPUTS : 

        ispace :: Flag indicating the spectral units
                  (0) Wavenumber (cm-1)
                  (1) Wavelength (um)
        wave(nwave) :: Wavelength or wavenumber array
        temp(ntemp) :: Temperature of the blackbody (K)

    OPTIONAL INPUTS:  none

    OUTPUTS : 

	    bb(nwave,ntemp) :: Planck function (W cm-2 sr-1 (cm-1)-1 or W cm-2 sr-1 um-1)
 
    CALLING SEQUENCE:

	    bb = planck(ispace,wave,temp)
 
    MODIFICATION HISTORY : Juan Alday (29/07/2021)

    """

    if(np.isscalar(temp)==True):  #Only one temperature value bb(nwave)
        ntemp = 0
    else:  #Several temperature values bb(nwave,ntemp)
        ntemp = len(temp)
        wave = np.repeat(wave[:, np.newaxis],ntemp,axis=1)

    c1 = 1.1911e-12
    c2 = 1.439
    if ispace==0:
        y = wave
        a = c1 * (y**3.)
    elif ispace==1:
        y = 1.0e4/wave
        a = c1 * (y**5.) / 1.0e4
    else:
        raise ValueError('error in planck :: ISPACE must be either 0 or 1')

    tmp = c2 * y / temp
    b = np.exp(tmp) - 1
    bb = a/b

    return bb

###############################################################################################
def planckg(ispace,wave,temp):
    """
    FUNCTION NAME : planckg()

    DESCRIPTION : Function to calculate the blackbody radiation given by the Planck function
                    as well as its derivative with respect to temperature

    INPUTS : 

        ispace :: Flag indicating the spectral units
                  (0) Wavenumber (cm-1)
                  (1) Wavelength (um)
        wave(nwave) :: Wavelength or wavenumber array
        temp :: Temperature of the blackbody (K)

    OPTIONAL INPUTS:  none

    OUTPUTS : 

	    bb(nwave) :: Planck function (W cm-2 sr-1 (cm-1)-1 or W cm-2 sr-1 um-1)
        dBdT(nwave) :: Temperature gradient (W cm-2 sr-1 (cm-1)-1 or W cm-2 sr-1 um-1)/K
 
    CALLING SEQUENCE:

	    bb,dBdT = planckg(ispace,wave,temp)
 
    MODIFICATION HISTORY : Juan Alday (29/07/2021)

    """

    c1 = 1.1911e-12
    c2 = 1.439
    if ispace==0:
        y = wave
        a = c1 * (y**3.)
        ap = c1 * c2 * (y**4.)/temp**2.
    elif ispace==1:
        y = 1.0e4/wave
        a = c1 * (y**5.) / 1.0e4
        ap = c1 * c2 * (y**6.) / 1.0e4 / temp**2.
    else:
        raise ValueError('error in planck :: ISPACE must be either 0 or 1')

    tmp = c2 * y / temp
    b = np.exp(tmp) - 1
    bb = a/b

    tmpp = c2 * y / temp
    bp = (np.exp(tmp) - 1.)**2.
    tp = np.exp(tmpp) * ap
    dBdT = tp/bp

    return bb,dBdT

##################################################################################################################
#Hapke BDRF calculations
##################################################################################################################
 
#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_BRDF(w,K,BS0,hs,BC0,hc,ROUGHNESS,G1,G2,F,i,e,phi):
    """
    Calculate the bidirectional-reflectance distribution function for a Hapke surface.
    The method used here is described in Hapke (2012): Theory of Reflectance and Emittance
    Spectroscopy, in chapter 12.3.1 (disk-resolved photometry)
    
    Inputs
    ______

    w(nwave) :: Single scattering albedo
    K(nwave) :: Porosity coefficient
    BS0(nwave), hs(nwave) :: Amplitude of opposition effect and width of the opposition surge
    BC0(nwave), hc(nwave) :: Amplitude of the coherent backscatter opposition effect and width of the backscatter function
    ROUGHNESS(nwave) :: Roughness mean slope angle
    G1(nwave),G2(nwave),F(nwave) :: Parameters describing the double Henyey-Greenstein phase function
    i(ntheta),e(ntheta),phi(ntheta) :: Incident, reflection and azimuth angle (degrees)
    
    Outputs
    _______
    
    BRDF(nwave,ntheta) :: Bidirectional reflectance
    
    """
    
    nwave = len(w)
    ntheta = len(i)
    
    BRDF = np.zeros((nwave,ntheta))
    
    for iwave in range(nwave):
        for itheta in range(ntheta):
            
            #print(i[itheta],e[itheta],phi[itheta])
            BRDF[iwave,itheta] = calc_Hapke_BRDFx(w[iwave],K[iwave],BS0[iwave],hs[iwave],BC0[iwave],hc[iwave],ROUGHNESS[iwave],\
                                                  G1[iwave],G2[iwave],F[iwave],i[itheta],e[itheta],phi[itheta])

    return BRDF

##################################################################################################################
    
#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_BRDFx(w,K,BS0,hs,BC0,hc,ROUGHNESS,G1,G2,F,i,e,phi_nemesis):
    """
    Calculate the bidirectional-reflectance distribution function for a Hapke surface.
    The method used here is described in Hapke (2012): Theory of Reflectance and Emittance
    Spectroscopy, in chapter 12.3.1 (disk-resolved photometry)
    
    Note that in the Hapke model the azimuth angle is defined opposite to NEMESIS.
    In NEMESIS forward scattering is phi=0 and backward scattering is phi=180.
    In Hapke forward scattering is phi=180 and backward scattering is phi=0.
    This is why we need to change the azimuth angle in the Hapke model to be consistent with NEMESIS.

    Inputs
    ______

    w :: Single scattering albedo
    K :: Porosity coefficient
    BS0, hs :: Amplitude of opposition effect and width of the opposition surge
    BC0, hc :: Amplitude of the coherent backscatter opposition effect and width of the backscatter function
    ROUGHNESS :: Roughness mean slope angle
    G1,G2,F :: Parameters describing the double Henyey-Greenstein phase function
    i,e,phi :: Incident, reflection and azimuth angle (degrees)
    
    Outputs
    _______
    
    BRDF :: Bidirectional reflectance
    
    """

    phi = 180. - phi_nemesis  #Changing the azimuth angle to be consistent with Hapke model

    if( (e>=90.) or (i>=90.) ):
    
        BRDF = 0.0
    
    else:
        
        #Calculating the cosine of the angl
        mu = np.cos(e/180.*np.pi)
        mu0 = np.cos(i/180.*np.pi)
        
        #Correcting the azimuth angle to be within 0-180 degrees
        if phi>180.:
            phix = 180. - (phi-180.)
        elif phi<0.:
            phix = -phi
        else:
            phix = phi
            
        #Calculating the phase angle
        cg = mu * mu0 + np.sqrt(1. - mu**2.) * np.sqrt(1. - mu0**2.) * np.cos(phix/180.*np.pi) 
        if cg>1.0:
            cg = 1.0
        if cg<0.0:
            cg = 0.0
        g = np.arccos(cg)/np.pi*180.   #Phase angle (degrees) (NTHETA)
        
        #Calculate some of the input parameters for the Hapke formalism
        gamma = np.sqrt(1. - w)
        r0 = (1. - gamma)/(1. + gamma)
        theta_bar = ROUGHNESS * (1. - r0)
        chi = 1./np.sqrt(1. + np.pi * np.tan(theta_bar/180.*np.pi)**2.)
        if np.abs(phix)==180.:
            fphi = 0.0
        else:
            fphi = np.exp(-2.*np.abs(np.tan(phix/2./180.*np.pi)))  #f(phi)

        #Calculating the E-functions
        E1e = calc_Hapke_E1(e,theta_bar)
        E2e = calc_Hapke_E2(e,theta_bar)
        E1i = calc_Hapke_E1(i,theta_bar)
        E2i = calc_Hapke_E2(i,theta_bar)

        #Calculating the nu functions
        nue = calc_Hapke_nu(e,theta_bar,E1e,E2e,chi)
        nui = calc_Hapke_nu(i,theta_bar,E1i,E2i,chi)
        
        #Calculating the effective incidence and reflection angles
        mu0eff, mueff = calc_Hapke_eff_angles(i,e,phix,theta_bar,E1e,E1i,E2e,E2i,chi)

        #Calculating the shadowing function S
        if i<=e:
            S = mueff/nue * mu0/nui * chi / (1.0 - fphi + fphi*chi*mu0/nui)
        else:
            S = mueff/nue * mu0/nui * chi / (1.0 - fphi + fphi*chi*mu/nue)

        #Calculating the shadow-hiding opposition function Bs
        Bs = BS0 / ( 1. + (1./hs) * np.tan( g/2./180.*np.pi) )
 
        #Calculating the backscatter anfular function Bc
        Bc = BC0 / ( 1. + (1.3 + K) * ( (1./hc*np.tan( g/2./180.*np.pi)) + (1./hc*np.tan( g/2./180.*np.pi))**2.0 ) )

        #Calculating the Ambartsumian–Chandrasekhar H function
        H0e = calc_Hapke_H(w,mu0eff/K,r0)
        He = calc_Hapke_H(w,mueff/K,r0)
         
        #Calculate phase function (double Henyey-Greenstein function)
        phase = calc_Hapke_hgphase(g,G1,G2,F)

        #Calculating the bidirectional reflectance
        r = K * w / (4.*np.pi) * mu0eff / (mu0eff + mueff) * ( phase*(1.+Bs) + (H0e*He-1.) ) * (1.+Bc) * S
        
        #Dividing the BRDF by the cosine of the incidence angle to account for the relation r(i,e,phi) = BRDF(i,e,phi) * cos(i)
        BRDF = r / np.cos(i/180.*np.pi)

    return BRDF


##################################################################################################################
    
#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_H(SGLALB,x,r0):
    """
    Calculate the Ambartsumian–Chandrasekhar H function of the Hapke formalism (Hapke, 2012; p. 333)

    Inputs
    ------

    SGLALB :: Single scattering albedo
    x :: Value at which the H-function must be evaluated
    r0 :: r0 parameter from the Hapke formalism

    Outputs
    -------

    @param H: 1D array or real scalar
        H function

    """

    H = 1.0 / ( 1.0 - SGLALB*x * (r0 + (1.0 - 2.0*r0*x)/2.0*np.log((1.0+x)/x)) )

    return H

##################################################################################################################

#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_thetabar(ROUGHNESS,r0):
    """
    Calculate the theta_bar parameter of the Hapke formalism (Hapke, 2012; p. 333)
    This parameter is the corrected roughness mean slope angle

    Inputs
    ------

    ROUGHNESS :: Roughness mean slope angle (degrees)
    r0 :: Diffusive reflectance

    Outputs
    -------

    theta_bar :: Corrected Roughness mean slope angle (degrees)

    """

    theta_bar = ROUGHNESS * (1.0 - r0)

    return theta_bar

##################################################################################################################

#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_gamma(SGLALB):
    """
    Calculate the gamma parameter of the Hapke formalism (Hapke, 2012; p. 333)
    This parameter is just a factor calculated from the albedo

    Inputs
    ------

    SGLALB :: Single scattering albedo

    Outputs
    -------

    gamma :: Gamma factor

    """

    gamma = np.sqrt(1.0 - SGLALB)

    return gamma

##################################################################################################################

#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_r0(gamma):
    """
    Calculate the r0 parameter of the Hapke formalism (Hapke, 2012; p. 333)
    This parameter is called the diffusive reflectance

    Inputs
    ------

    gamma :: Gamma factor

    Outputs
    -------

    r0 :: Diffusive reflectance

    """

    r0 = (1.0 - gamma)/(1.0 + gamma)

    return r0

##################################################################################################################


#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_eff_angles(i,e,phi,theta_bar,E1e,E1i,E2e,E2i,chi):
    """
    Calculate the effective incidence and reflection angles 

    Inputs
    ------

    i :: Incidence angle (degrees)  
    e :: Reflection angle (degrees)
    phi :: Azimuth angle (degrees)
    theta_bar :: Corrected roughness mean slope angle (degrees)
    E1e,E1i,E2e,E2i,chi :: Several different coefficients from the Hapke formalism         

    Outputs
    -------

    mu0_eff :: Cosine of the effective incidence angle
    mu_eff :: Cosine of the effective reflection angle

    """

    #Calculating some initial parameters
    irad = i / 180. * np.pi  
    erad = e / 180. * np.pi
    phirad = phi / 180. * np.pi 
    tbarrad = theta_bar / 180. * np.pi

    #There are two possible cases
    if i<=e:

        mu0eff = chi * ( np.cos(irad) + np.sin(irad) * np.tan(tbarrad) * \
            (np.cos(phirad) * E2e + np.sin(phirad/2.)**2. *E2i) / (2.0 - E1e - phirad/np.pi*E1i)  )

        mueff = chi * ( np.cos(erad) + np.sin(erad) * np.tan(tbarrad) * \
            (E2e - np.sin(phirad/2.)**2. *E2i) / (2.0 - E1e - phirad/np.pi*E1i)  )

    elif i>e:

        mu0eff = chi * ( np.cos(irad) + np.sin(irad) * np.tan(tbarrad) * \
            (E2i - np.sin(phirad/2.)**2. *E2e) / (2.0 - E1i - phirad/np.pi*E1e)  )

        mueff = chi * ( np.cos(erad) + np.sin(erad) * np.tan(tbarrad) * \
            (np.cos(phirad) * E2i + np.sin(phirad/2.)**2. *E2e) / (2.0 - E1i - phirad/np.pi*E1e)  )


    return mu0eff, mueff

##################################################################################################################

#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_nu(x,theta_bar,E1x,E2x,chi):
    """
    Calculate the nu function from the Hapke formalism (Hapke 2012 p.333) 

    Inputs
    ------

    x :: Incidence or reflection angles (degrees)
    theta_bar :: Corrected roughness mean slope angle (degrees)
    E1x,E1x,chi :: Several different coefficients from the Hapke formalism (evaluated at the angle x)       

    Outputs
    -------

    nu :: Nu parameter from the Hapke formalism

    """

    #Calculating some initial parameters
    xrad = x / 180. * np.pi
    tbarrad = theta_bar / 180. * np.pi

    nu = chi * ( np.cos(xrad) + np.sin(xrad) * np.tan(tbarrad) * \
        (E2x) / (2.0 - E1x)  )

    return nu


##################################################################################################################

#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_E1(x,theta_bar):
    """
    Calculate the E1 function of the Hapke formalism (Hapke, 2012; p. 333)

    Inputs
    ------

    x :: Angle (degrees)
    theta_bar :: Mean slope angle (degrees)

    Outputs
    -------

    E1 :: Parameter E1 in the Hapke formalism

    """

    if( (theta_bar==0.0) or (x==0.0) ):
        E1 = 0.0
    else:
        E1 = np.exp(-2.0/np.pi * 1.0/np.tan(theta_bar/180.*np.pi) * 1./np.tan(x/180.*np.pi))

    return E1


##################################################################################################################

#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_E2(x,theta_bar):
    """
    Calculate the E2 function of the Hapke formalism (Hapke, 2012; p. 333)

    Inputs
    ------

    x :: Angle (degrees)
    theta_bar :: Mean slope angle (degrees)

    Outputs
    -------

    E2 :: Parameter E2 in the Hapke formalism

    """

    if( (theta_bar==0.0) or (x==0.0) ):
        E2 = 0.0
    else:
        E2 = np.exp(-1.0/np.pi * 1.0/np.tan(theta_bar/180.*np.pi)**2. * 1./np.tan(x/180.*np.pi)**2.)

    return E2

##################################################################################################################

#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_chi(theta_bar):
    """
    Calculate the chi function of the Hapke formalism (Hapke, 2012; p. 333)

    Inputs
    ------

    theta_bar :: Corrected roughness mean slope angle (degrees)

    Outputs
    -------

    chi :: Parameter chi in the Hapke formalism

    """

    chi = 1./np.sqrt(1.0 + np.pi * np.tan(theta_bar/180.*np.pi)**2.)

    return chi


##################################################################################################################

#@njit(fastmath=True)
@conditional_jit(nopython=True)
def calc_Hapke_hgphase(Theta,G1,G2,F):
    """
    Calculate the phase function at Theta angles given the double Henyey-Greenstein parameters
    
    Inputs
    ------
    
    Theta :: Scattering angle (degrees)
    
    Outputs
    -------

    phase :: Phase function evaluated at Theta
    """

    t1 = (1.-G1**2.)/(1. - 2.*G1*np.cos(Theta/180.*np.pi) + G1**2.)**1.5
    t2 = (1.-G2**2.)/(1. - 2.*G2*np.cos(Theta/180.*np.pi) + G2**2.)**1.5
    
    phase = F * t1 + (1.0 - F) * t2

    return phase


##################################################################################################################
#Oren & Nayar BDRF calculations
##################################################################################################################

@jit(nopython=True)
def calc_OrenNayar_BRDF(A,ROUGHNESS,i,e,phi):
    """
    Calculate the bidirectional-reflectance distribution function for a surface following the
    method described by Oren & Nayar (1994). This method is a generalisation of the Lambertian model
    for rough surfaces.

    Inputs
    ______

    A(nwave) :: Lambertian albedo
    ROUGHNESS(nwave) :: Roughness parameter (degrees)
    i(ntheta),e(ntheta),phi(ntheta) :: Incident, reflection and azimuth angle (degrees)
    
    Outputs
    _______
    
    BRDF(nwave,ntheta) :: Bidirectional reflectance
    
    """
    
    nwave = len(A)
    ntheta = len(i)
    
    BRDF = np.zeros((nwave,ntheta))
    
    for iwave in range(nwave):
        for itheta in range(ntheta):
            
            BRDF[iwave,itheta] = calc_OrenNayar_BRDFx(A[iwave],ROUGHNESS[iwave],i[itheta],e[itheta],phi[itheta])
            
    return BRDF


@jit(nopython=True)
def calc_OrenNayar_BRDFx(A,ROUGHNESS,i,e,phi):
    """
    Calculate the bidirectional-reflectance distribution function for a surface following the
    method described by Oren & Nayar (1994). This method is a generalisation of the Lambertian model
    for surface with roughness

    Inputs
    ______

    A :: Lambertian albedo
    ROUGHNESS :: Roughness parameter (degrees)
    i,e,phi :: Incident, reflection and azimuth angle (degrees)
    
    Outputs
    _______
    
    BRDF :: Bidirectional reflectance
    
    """
    
    #Converting angles and roughness to radians
    irad = i / 180. * np.pi
    erad = e / 180. * np.pi
    phirad = phi / 180. * np.pi
    sigma = ROUGHNESS / 180. * np.pi
    
    #Calculating initial parameters
    alpha = max([irad,erad])
    beta = min([irad,erad])
    
    #Calculating the C parameters
    C1 = 1.0 - 0.5 * (sigma)**2. / ( (sigma)**2. + 0.33 )
    
    C2 = 0.45 * (sigma)**2. / ( (sigma)**2. + 0.09 )
    if np.cos(phirad)>=0:
        C2 *= np.sin(alpha)
    else:
        C2 *= (np.sin(alpha) - (2. * beta/ np.pi)**3. )
    
    C3 = 0.125 * sigma**2. / ( sigma**2. + 0.09 ) * (4.*alpha*beta/np.pi**2.)**2.
    
    #Calculating the L terms
    BRDF1 = A/np.pi * ( C1 + np.cos(phirad) * C2 * np.tan(beta) + (1. - np.abs(np.cos(phirad))) * C3 * np.tan((alpha+beta)/2.) )
    BRDF2 = 0.17 * A**2. /np.pi * sigma**2. /  ( sigma**2. + 0.13 ) * (1.0 - np.cos(phirad) * (2.*beta/np.pi)**2. )

    BRDF = BRDF1 + BRDF2
    
    return BRDF


from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

setup(
    name="archnemesis",
    version="1.0.6",
    author="Juan Alday",
    description="Python implementation of the NEMESIS radiative transfer and retrieval code",
    long_description=long_description,
    long_description_content_type="text/markdown",  # important for Markdown rendering
    url="https://github.com/juanaldayparejo/archnemesis-dist",
    project_urls={
        "Documentation": "https://archnemesis.readthedocs.io",
        "Source": "https://github.com/juanaldayparejo/archnemesis-dist",
        "Tracker": "https://github.com/juanaldayparejo/archnemesis-dist/issues",
        "DockerHub": "https://hub.docker.com/r/juanaldayparejo/archnemesis",
    },
    #packages=["archnemesis"],
    packages=find_packages(),  #automatically include all subpackages like archnemesis.cfg
    install_requires=[
      'numpy',
      'matplotlib',
      'numba>=0.57.0',
      'scipy',
      'pymultinest',
      'cdsapi',
      'joblib',
      'h5py',
      'basemap',
      'pytest',
      'corner',
      'typing_extensions'
    ],
    extras_require={
        'grib': ['pygrib'],
        'docs': ['sphinx', 'sphinx_rtd_theme'],
        'spectroscopy': ['hitran-api'], # The HITRAN api module "hapi" is called "hitran-api" on pypi
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)
